# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team
# Copyright (C) 2009-2013 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function
import cairo
import math
import logging

import gui.drawutils
from lib.pycompat import xrange

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GdkPixbuf

logger = logging.getLogger(__name__)

# Absolute minimum size
BRUSH_CURSOR_MIN_SIZE = 3

# Cursor style constants
BRUSH_CURSOR_STYLE_NORMAL = 0
BRUSH_CURSOR_STYLE_ERASER = 1
BRUSH_CURSOR_STYLE_LOCK_ALPHA = 2
BRUSH_CURSOR_STYLE_COLORIZE = 3

# cache only the last cursor
last_cursor_info = None
last_cursor = None
max_cursor_size = None


class Name:
    """Cursor name constants

    These correspond to the names of small (22x22px) png files in the
    installation's pixmaps directory.

    """

    ARROW = "cursor_arrow"
    ARROW_MOVE = "cursor_arrow_move"
    PENCIL = "cursor_pencil"
    HAND_OPEN = "cursor_hand_open"
    HAND_CLOSED = "cursor_hand_closed"
    CROSSHAIR_OPEN = "cursor_crosshair_open"
    CROSSHAIR_OPEN_PRECISE = "cursor_crosshair_precise_open"
    CROSSHAIR_CLOSED = "cursor_crosshair_closed"
    INVERTED_CROSSHAIR_DIAGONAL = "cursor_crosshair_diagonal"
    MOVE_WEST_OR_EAST = "cursor_move_w_e"
    MOVE_NORTHWEST_OR_SOUTHEAST = "cursor_move_nw_se"
    MOVE_NORTH_OR_SOUTH = "cursor_move_n_s"
    MOVE_NORTHEAST_OR_SOUTHWEST = "cursor_move_ne_sw"
    FORBIDDEN_EVERYWHERE = "cursor_forbidden_everywhere"
    ARROW_FORBIDDEN = "cursor_arrow_forbidden"
    REMOVE = "cursor_remove"
    ADD = "cursor_add"
    PICKER = "cursor_color_picker"
    ERASER = "cursor_eraser"
    ALPHA_LOCK = "cursor_alpha_lock"
    COLORIZE = "cursor_colorize"


def get_brush_cursor(radius, style, prefs={}):
    """Get a cursor for use with a brush of a particular size and type

    :param float radius: Radius of the ring
    :param int style: A cursor style constant (for now, see the source)
    :param dict prefs: User preferences dict

    """
    global last_cursor, last_cursor_info, max_cursor_size

    display = Gdk.Display.get_default()
    if not max_cursor_size:
        max_cursor_size = max(display.get_maximal_cursor_size())
    d = int(radius*2)
    min_size = max(prefs.get("cursor.freehand.min_size", 4),
                   BRUSH_CURSOR_MIN_SIZE)
    if d < min_size:
        d = min_size
    if d+1 > max_cursor_size:
        d = max_cursor_size-1
    cursor_info = (d, style, min_size)
    if cursor_info != last_cursor_info:
        last_cursor_info = cursor_info
        surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, d+1, d+1)
        cr = cairo.Context(surf)
        cr.set_source_rgba(1, 1, 1, 0)
        cr.paint()
        draw_brush_cursor(cr, d, style, prefs)
        surf.flush()

        # Calculate hotspot. Zero means topmost or leftmost. Cursors with an
        # even pixel diameter are interesting because they can never be
        # perfectly centred on their hotspot. Rounding down and not up may be
        # more "arrow-cursor" like in this case.
        #
        # NOTE: is it worth adjusting the freehand drawing code to add half a
        # pixel to the position passed on to brushlib for the even case?
        hot_x = hot_y = int(d // 2)

        pixbuf = _image_surface_to_pixbuf(surf)
        last_cursor = Gdk.Cursor.new_from_pixbuf(display, pixbuf,
                                                 hot_x, hot_y,)
    return last_cursor


def _image_surface_to_pixbuf(surf):
    """Convert a Cairo surface to a GdkPixbuf"""
    w = surf.get_width()
    h = surf.get_height()
    return Gdk.pixbuf_get_from_surface(surf, 0, 0, w, h)


def draw_brush_cursor(cr, d, style=BRUSH_CURSOR_STYLE_NORMAL, prefs={}):
    """Draw a brush cursor into a Cairo context

    :param cr: A Cairo context
    :param d: The diameter of the resultant brush cursor
    :param int style: A cursor style constant (for now, see the source)
    :param dict prefs: User preferences dict

    The Cairo context is assumed to have width and height `d`+1, at
    least.

    """

    # The cursor consists of an inner circle drawn over an outer one. The
    # inmost edge of the inner ring, and the outmode edge of the outer ring are
    # pixel-edge aligned. If vertical bars represent pixel edges, then
    # conceptually,
    #
    #   |<--------------    Integer radius
    #   |OOOOOO             Outer ring pixels
    #   |------->|          Integer inset
    #       IIIII|          Inner ring pixels
    #
    # Which results in the cleanest possible edges inside and outside. Note the
    # overlap, and that line widths don't have to be integers.

    # Outer and inner line widths
    width1 = float(prefs.get("cursor.freehand.outer_line_width", 1.25))
    width2 = float(prefs.get("cursor.freehand.inner_line_width", 1.25))
    inset = int(prefs.get("cursor.freehand.inner_line_inset", 2))

    # Colors
    col_bg = tuple(prefs.get(
        "cursor.freehand.outer_line_color",
        (0, 0, 0, 1),
    ))
    col_fg = tuple(prefs.get(
        "cursor.freehand.inner_line_color",
        (1, 1, 1, 0.75),
    ))

    # Cursor style
    arcs = []
    if style == BRUSH_CURSOR_STYLE_ERASER:
        # divide into eighths, alternating on and off
        k = math.pi / 4
        k2 = k/2
        arcs.append((k2, k2+k))
        arcs.append((k2+2*k, k2+3*k))
        arcs.append((k2+4*k, k2+5*k))
        arcs.append((k2+6*k, k2+7*k))
    elif style == BRUSH_CURSOR_STYLE_LOCK_ALPHA:
        # same thing, but the two side voids are filled
        k = math.pi/4
        k2 = k/2
        arcs.append((k2+6*k, k2+k))
        arcs.append((k2+2*k, k2+5*k))
    elif style == BRUSH_CURSOR_STYLE_COLORIZE:
        # same as lock-alpha, but with the voids turned through 90 degrees
        k = math.pi/4
        k2 = k/2
        arcs.append((k2, k2+3*k))
        arcs.append((k2+4*k, k2+7*k))
    else:
        # Regular drawing mode
        arcs.append((0, 2*math.pi))

    # Pick centre to ensure pixel alignedness for the outer edge of the
    # black outline.
    if d % 2 == 0:
        r0 = int(d // 2)
    else:
        r0 = int(d // 2) + 0.5
    cx = cy = r0

    # Outer "bg" line.
    cr.set_line_cap(cairo.LINE_CAP_BUTT)
    cr.set_source_rgba(*col_bg)
    cr.set_line_width(width1)
    r = r0 - (width1 / 2.0)
    for a1, a2 in arcs:
        cr.new_sub_path()
        cr.arc(cx, cy, r, a1, a2)
    cr.stroke()

    # Inner line: also pixel aligned, but to its inner edge.
    cr.set_line_cap(cairo.LINE_CAP_ROUND)
    cr.set_source_rgba(*col_fg)
    cr.set_line_width(width2)
    r = r0 - inset + (width2 / 2.0)
    for a1, a2 in arcs:
        cr.new_sub_path()
        cr.arc(cx, cy, r, a1, a2)
    cr.stroke()


class CustomCursorMaker (object):
    """Factory and cache of custom cursors for actions."""

    CURSOR_HOTSPOTS = {
        Name.ARROW: (1, 1),
        Name.ARROW_MOVE: (1, 1),
        Name.PENCIL: (7, 22),
        Name.HAND_OPEN: (11, 12),
        Name.HAND_CLOSED: (11, 12),
        Name.CROSSHAIR_OPEN: (11, 11),
        Name.CROSSHAIR_CLOSED: (11, 11),
        Name.CROSSHAIR_OPEN_PRECISE: (12, 11),
        Name.INVERTED_CROSSHAIR_DIAGONAL: (11, 11),
        Name.MOVE_WEST_OR_EAST: (11, 11),
        Name.MOVE_NORTHWEST_OR_SOUTHEAST: (11, 11),
        Name.MOVE_NORTH_OR_SOUTH: (11, 11),
        Name.MOVE_NORTHEAST_OR_SOUTHWEST: (11, 11),
        Name.FORBIDDEN_EVERYWHERE: (11, 11),
        Name.ARROW_FORBIDDEN: (7, 4),
        Name.REMOVE: (11, 11),
        Name.ADD: (11, 11),
        Name.PICKER: (3, 15),
        Name.ERASER: (12, 11),
        Name.ALPHA_LOCK: (12, 11),
        Name.COLORIZE: (12, 11),
    }

    def __init__(self, app):
        object.__init__(self)
        self.app = app
        self.cache = {}

    def _get_overlay_cursor(self, icon_pixbuf, cursor_name=Name.ARROW):
        """Returns an overlay cursor. Not cached.

        :param icon_pixbuf: a GdkPixbuf.Pixbuf containing a small (~22px)
           image, or None
        :param cursor_name: name of a pixmaps/ cursor image to use for the
           pointer part, minus the .png

        The overlay icon will be overlaid to the bottom and right of the
        returned cursor image.

        """

        pointer_pixbuf = getattr(self.app.pixmaps, cursor_name)
        pointer_w = pointer_pixbuf.get_width()
        pointer_h = pointer_pixbuf.get_height()
        hot_x, hot_y = self.CURSOR_HOTSPOTS.get(cursor_name, (None, None))
        if hot_x is None:
            hot_x = 1
            hot_y = 1

        cursor_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True,
                                             8, 32, 32)
        cursor_pixbuf.fill(0x00000000)

        pointer_pixbuf.composite(
            cursor_pixbuf, 0, 0, pointer_w, pointer_h, 0, 0, 1, 1,
            GdkPixbuf.InterpType.NEAREST, 255
        )
        if icon_pixbuf is not None:
            icon_w = icon_pixbuf.get_width()
            icon_h = icon_pixbuf.get_height()
            icon_x = 32 - icon_w
            icon_y = 32 - icon_h
            icon_pixbuf.composite(
                cursor_pixbuf, icon_x, icon_y, icon_w, icon_h,
                icon_x, icon_y, 1, 1, GdkPixbuf.InterpType.NEAREST, 255
            )

        display = self.app.drawWindow.get_display()
        cursor = Gdk.Cursor.new_from_pixbuf(display, cursor_pixbuf,
                                            hot_x, hot_y)
        return cursor

    def get_freehand_cursor(self, cursor_name=Name.CROSSHAIR_OPEN_PRECISE):
        """Returns a cursor for the current app.brush. Cached.

        :param cursor_name: name of a pixmaps/ image to use, minus the .png

        An icon for the brush blend mode will be overlaid to the bottom and
        right of the cursor image.

        """
        # Pick an icon
        if self.app.brush.is_eraser():
            icon_name = "mypaint-eraser-symbolic"
        elif self.app.brush.is_alpha_locked():
            icon_name = "mypaint-lock-alpha-symbolic"
        elif self.app.brush.is_colorize():
            icon_name = "mypaint-colorize-symbolic"
        else:
            icon_name = None
        return self.get_icon_cursor(icon_name, cursor_name)

    def get_action_cursor(self, action_name, cursor_name=Name.ARROW):
        """Returns an overlay cursor for a named action. Cached.

        :param action_name: the name of a GtkAction defined in resources.xml
        :param cursor_name: name of a pixmaps/ image to use, minus the .png

        The action's icon will be overlaid at a small size to the bottom and
        right of the cursor image.

        """
        # Find a small action icon for the overlay
        action = self.app.find_action(action_name)
        if action is None:
            return Gdk.Cursor.new(Gdk.CursorType.ARROW)
        icon_name = action.get_icon_name()
        if icon_name is None:
            return Gdk.Cursor.new(Gdk.CursorType.ARROW)
        return self.get_icon_cursor(icon_name, cursor_name)

    def get_icon_cursor(self, icon_name, cursor_name=Name.ARROW):
        """Returns an overlay cursor for a named icon. Cached.

        :param icon_name: themed icon system name.
        :param cursor_name: name of a pixmaps/ image to use, minus the .png

        The icon will be overlaid at a small size to the bottom and right of
        the cursor image.

        """

        # Return from cache, if we have an entry
        cache_key = ("actions", icon_name, cursor_name)
        if cache_key in self.cache:
            return self.cache[cache_key]

        if icon_name is not None:
            if "symbolic" in icon_name:
                icon_pixbuf = gui.drawutils.load_symbolic_icon(
                    icon_name, 18,
                    fg=(1, 1, 1, 1),
                    outline=(0, 0, 0, 1),
                )
            else:
                # Look up icon via the user's current theme
                icon_theme = Gtk.IconTheme.get_default()
                size_range = [Gtk.IconSize.SMALL_TOOLBAR, Gtk.IconSize.MENU]
                for icon_size in size_range:
                    valid, width, height = Gtk.icon_size_lookup(icon_size)
                    if not valid:
                        continue
                    size = min(width, height)
                    if size > 24:
                        continue
                    flags = 0
                    icon_pixbuf = icon_theme.load_icon(icon_name, size, flags)
                    if icon_pixbuf:
                        break
                if not icon_pixbuf:
                    logger.warning(
                        "Can't find icon %r for cursor. Search path: %r",
                        icon_name,
                        icon_theme.get_search_path(),
                    )
        else:
            icon_pixbuf = None

        # Build cursor
        cursor = self._get_overlay_cursor(icon_pixbuf, cursor_name)

        # Cache and return
        self.cache[cache_key] = cursor
        return cursor


def get_move_cursor_name_for_angle(theta):
    """Cursor name to use for a motion in a given direction"""
    while theta < 2*math.pi:
        theta += 2*math.pi
    theta %= 2*math.pi
    assert theta >= 0
    assert theta < 2*math.pi
    cursor_strs = [
        (1, Name.MOVE_WEST_OR_EAST),
        (3, Name.MOVE_NORTHWEST_OR_SOUTHEAST),
        (5, Name.MOVE_NORTH_OR_SOUTH),
        (7, Name.MOVE_NORTHEAST_OR_SOUTHWEST),
        (9, Name.MOVE_WEST_OR_EAST),
        (11, Name.MOVE_NORTHWEST_OR_SOUTHEAST),
        (13, Name.MOVE_NORTH_OR_SOUTH),
        (15, Name.MOVE_NORTHEAST_OR_SOUTHWEST),
        (17, Name.MOVE_WEST_OR_EAST),
    ]
    cursor_str = None
    for i, s in cursor_strs:
        if theta < i*(2.0/16)*math.pi:
            cursor_str = s
            break
    assert cursor_str is not None
    return cursor_str


## Interactive testing

if __name__ == '__main__':
    from random import randint
    win = Gtk.Window()
    win.set_title("cursor test")

    min_size = 5
    max_size = 64
    nsteps = 8
    w = nsteps * max_size
    h = 4 * max_size
    surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
    cr = cairo.Context(surf)
    cr.set_source_rgb(.7, .7, .7)
    cr.paint()

    for style in xrange(4):
        col = 0
        for size in xrange(min_size, max_size + 1,
                           (max_size - min_size) // nsteps):
            cr.save()
            y = (style * max_size) + ((max_size - size)/2)
            x = (col * max_size) + ((max_size - size)/2)
            cr.translate(x, y)
            draw_brush_cursor(cr, size, style)
            cr.restore()
            col += 1
    pixbuf = _image_surface_to_pixbuf(surf)
    image = Gtk.Image()
    image.set_from_pixbuf(pixbuf)
    image.set_size_request(w, h)

    display = Gdk.Display.get_default()
    max_size = max(display.get_maximal_cursor_size())
    num_styles = 4
    style = 0

    def _enter_cb(widget, event):
        global style, max_size
        r = randint(3, max_size // 2)
        print("DEBUG: radius=%s, style=%s" % (r, style))
        cursor = get_brush_cursor(r, style)
        widget.get_window().set_cursor(cursor)
        style += 1
        if style >= num_styles:
            style = 0
    win.connect("enter-notify-event", _enter_cb)
    win.add(image)
    win.connect("destroy", lambda *a: Gtk.main_quit())
    win.show_all()
    Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Overlays for TDWs showing information about the TDW state."""


## Imports

from __future__ import division, print_function
from math import pi
from gettext import gettext as _

from lib.gibindings import Pango
from lib.gibindings import PangoCairo
from lib.gibindings import GLib
import cairo

from lib.helpers import clamp
import gui.style


## Base classes and utils

class Overlay (object):
    """Base class/interface for objects which paint things over a TDW."""

    def paint(self, cr):
        """Paint information onto a TiledDrawWidget.

        The drawing interface is very simple. `cr` is a Cairo context in either
        display coordinates or model coordinates: which one you get depends on
        which list the Overlay is appended to on its tdw.
        """
        pass


class FadingOverlay (Overlay):
    """Base class for temporary overlays which fade to alpha over a short time
    """

    # Overridable animation controls
    fade_fps = 20  #: Nominal frames per second
    fade_duration = 1.5  #: Time for fading entirely to zero, in seconds

    # Animation and redrawing state
    tdw = None
    alpha = 1.0
    __area = None
    __anim_srcid = None

    def __init__(self, doc):
        Overlay.__init__(self)
        self.tdw = doc.tdw

    def paint(self, cr):
        """Repaint the overlay and start animating if necessary.

        Individual frames are handled by `paint_frame()`.
        """
        if self.overlay_changed():
            self.alpha = 1.0
        elif self.alpha <= 0:
            # No need to draw anything.
            return
        self.__restart_anim_if_needed()
        self.__area = self.paint_frame(cr)

    def anim_cb(self):
        """Animation callback.

        Each step fades the alpha multiplier slightly and invalidates the area
        last painted.
        """
        self.alpha -= 1 / (self.fade_fps * self.fade_duration)
        self.alpha = clamp(self.alpha, 0.0, 1.0)

        if self.__area:
            self.tdw.queue_draw_area(*self.__area)
        if self.alpha <= 0.0:
            self.__anim_srcid = None
            return False
        else:
            return True

    def __restart_anim_if_needed(self):
        """Restart if not currently running, without changing the alpha.
        """
        if self.__anim_srcid is None:
            delay = int(1000 // self.fade_fps)
            self.__anim_srcid = GLib.timeout_add(delay, self.anim_cb)

    def stop_anim(self):
        """Stops the animation after the next frame is drawn.
        """
        self.alpha = 0.0

    def start_anim(self):
        """Restarts the animation, setting alpha to 1.
        """
        self.alpha = 1.0
        self.__restart_anim_if_needed()

    def paint_frame(self, cr):
        """Paint a single frame.
        """
        raise NotImplementedError

    def overlay_changed(self):
        """Return true if the overlay has changed.

        This virtual method is called by paint() to determine whether the
        alpha should be reset to 1.0 and the fade begun anew.
        """
        raise NotImplementedError


def rounded_box(cr, x, y, w, h, r):
    """Paint a rounded box path into a Cairo context.

    The position is given by `x` and `y`, and the size by `w` and `h`. The
    cornders are of radius `r`, and must be smaller than half the minimum
    dimension. The path is created as a new, closed subpath.
    """
    assert r <= min(w, h) / 2
    cr.new_sub_path()
    cr.arc(x+r, y+r, r, pi, pi*1.5)
    cr.line_to(x+w-r, y)
    cr.arc(x+w-r, y+r, r, pi*1.5, pi*2)
    cr.line_to(x+w, y+h-r)
    cr.arc(x+w-r, y+h-r, r, 0, pi*0.5)
    cr.line_to(x+r, y+h)
    cr.arc(x+r, y+h-r, r, pi*0.5, pi)
    cr.close_path()


## Minor builtin overlays


class ScaleOverlay (FadingOverlay):
    """Overlays its TDW's current zoom, fading to transparent.

    The animation is started by the normal full canvas repaint which happens
    after the scale changes.
    """

    vmargin = 6
    hmargin = 12
    padding = 6
    shown_scale = None

    def overlay_changed(self):
        return self.tdw.scale != self.shown_scale

    def paint_frame(self, cr):
        self.shown_scale = self.tdw.scale
        text = _("Zoom: %.01f%%") % (100*self.shown_scale)
        layout = self.tdw.create_pango_layout(text)

        # Set a bold font
        font = layout.get_font_description()
        if font is None:  # inherited from context
            font = layout.get_context().get_font_description()
            font = font.copy()
        font.set_weight(Pango.Weight.BOLD)
        layout.set_font_description(font)

        # General dimensions
        alloc = self.tdw.get_allocation()
        lw, lh = layout.get_pixel_size()

        # Background rectangle
        hm = self.hmargin
        vm = self.hmargin
        w = alloc.width
        p = self.padding
        area = bx, by, bw, bh = w-lw-hm-p-p, vm, lw+p+p, lh+p+p
        rounded_box(cr, bx, by, bw, bh, p)
        rgba = list(gui.style.TRANSIENT_INFO_BG_RGBA)
        rgba[3] *= self.alpha
        cr.set_source_rgba(*rgba)
        cr.fill()

        # Text
        cr.translate(w-lw-hm-p, vm+p)
        rgba = list(gui.style.TRANSIENT_INFO_RGBA)
        rgba[3] *= self.alpha
        cr.set_source_rgba(*rgba)
        PangoCairo.show_layout(cr, layout)

        # Where to invalidate
        return area


class LastPaintPosOverlay (FadingOverlay):
    """Displays the last painting position after a stroke has finished.

    Not especially useful, but serves as an example of how to drive an overlay
    from user input events.
    """

    inner_line_rgba = gui.style.TRANSIENT_INFO_RGBA
    inner_line_width = 6
    outer_line_rgba = gui.style.TRANSIENT_INFO_BG_RGBA
    outer_line_width = 8
    radius = 4.0

    def __init__(self, doc):
        FadingOverlay.__init__(self, doc)
        doc.input_stroke_started += self.input_stroke_started
        doc.input_stroke_ended += self.input_stroke_ended
        self.current_marker_pos = None
        self.in_input_stroke = False

    def input_stroke_started(self, doc, event):
        self.in_input_stroke = True
        if self.current_marker_pos is None:
            return
        # Clear the current marker
        model_x, model_y = self.current_marker_pos
        x, y = self.tdw.model_to_display(model_x, model_y)
        area = self._calc_area(x, y)
        self.tdw.queue_draw_area(*area)
        self.current_marker_pos = None
        self.stop_anim()

    def input_stroke_ended(self, doc, event):
        self.in_input_stroke = False
        if self.tdw.last_painting_pos is None:
            return
        # Record the new marker position
        model_x, model_y = self.tdw.last_painting_pos
        x, y = self.tdw.model_to_display(model_x, model_y)
        area = self._calc_area(x, y)
        self.tdw.queue_draw_area(*area)
        self.current_marker_pos = model_x, model_y
        self.start_anim()

    def overlay_changed(self):
        return False

    def _calc_area(self, x, y):
        r = self.radius
        lw = max(self.inner_line_width, self.outer_line_width)
        return (int(x-r-lw), int(y-r-lw), int(2*(r+lw)), int(2*(r+lw)))

    def paint_frame(self, cr):
        if self.in_input_stroke:
            return
        if self.current_marker_pos is None:
            return
        x, y = self.tdw.model_to_display(*self.current_marker_pos)
        area = self._calc_area(x, y)
        x = int(x) + 0.5
        y = int(y) + 0.5
        r = self.radius
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        rgba = list(self.outer_line_rgba)
        rgba[3] *= self.alpha
        cr.set_source_rgba(*rgba)
        cr.set_line_width(self.outer_line_width)
        cr.move_to(x-r, y-r)
        cr.line_to(x+r, y+r)
        cr.move_to(x+r, y-r)
        cr.line_to(x-r, y+r)
        cr.stroke_preserve()
        rgba = list(self.inner_line_rgba)
        rgba[3] *= self.alpha
        cr.set_source_rgba(*rgba)
        cr.set_line_width(self.inner_line_width)
        cr.stroke()
        return area



# This file is part of MyPaint.
# Copyright (C) 2007-2018 by the MyPaint Development Team
# Copyright (C) 2007 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function
import logging
import copy
import math
import json

from lib import mypaintlib
from lib import helpers
from lib import brushsettings
from lib.eotf import eotf
from lib.pycompat import unicode
from lib.pycompat import PY3

if PY3:
    from urllib.parse import quote_from_bytes as url_quote
    from urllib.parse import unquote_to_bytes as url_unquote
else:
    from urllib import quote as url_quote
    from urllib import unquote as url_unquote

logger = logging.getLogger(__name__)


# Module constants:

STRING_VALUE_SETTINGS = set((
    "parent_brush_name",
    "group",  # Possibly obsolete group field (replaced by order.conf?)
    "comment",  # MyPaint uses this to explanation what the file is
    "notes",  # Brush developer's notes field, multiline
    "description",  # Short, user-facing description field, single line
))
OLDFORMAT_BRUSHFILE_VERSION = 2

BRUSH_SETTINGS = set([s.cname for s in brushsettings.settings])
ALL_SETTINGS = BRUSH_SETTINGS.union(STRING_VALUE_SETTINGS)

_BRUSHINFO_MATCH_IGNORES = [
    "color_h", "color_s", "color_v",
    "parent_brush_name",
]


# Helper funcs for quoting and unquoting:

def brushinfo_quote(string):
    """Quote a string for serialisation of brushes.

    >>> brushinfo_quote(u'foo') == b'foo'
    True
    >>> brushinfo_quote(u'foo/bar blah') == b'foo%2Fbar%20blah'
    True
    >>> expected = b'Have%20a%20nice%20day%20%E2%98%BA'
    >>> brushinfo_quote(u'Have a nice day \u263A') == expected
    True

    """
    string = unicode(string)
    u8bytes = string.encode("utf-8")
    return url_quote(u8bytes, safe='').encode("ascii")


def brushinfo_unquote(quoted):
    """Unquote a serialised string value from a brush field.

    >>> brushinfo_unquote(b"foo") == u'foo'
    True
    >>> brushinfo_unquote(b"foo%2fbar%20blah") == u'foo/bar blah'
    True
    >>> expected = u'Have a nice day \u263A'
    >>> brushinfo_unquote(b'Have%20a%20nice%20day%20%E2%98%BA') == expected
    True

    """
    if not isinstance(quoted, bytes):
        raise ValueError("Cann")
    u8bytes = url_unquote(quoted)
    return unicode(u8bytes.decode("utf-8"))


# Exceptions raised during brush parsing:

class ParseError (Exception):
    pass


class Obsolete (ParseError):
    pass


# Helper functions for parsing the old brush format:

def _oldfmt_parse_value(rawvalue, cname, version):
    """Parses a raw setting value.

    This code handles a format that changed over time, so the
    parse is for a given setting name and brushfile version.

    """
    if cname in STRING_VALUE_SETTINGS:
        string = brushinfo_unquote(rawvalue)
        return [(cname, string)]
    elif version <= 1 and cname == 'color':
        rgb = [int(c) / 255.0 for c in rawvalue.split(" ")]
        h, s, v = helpers.rgb_to_hsv(*rgb)
        return [
            ('color_h', [h, {}]),
            ('color_s', [s, {}]),
            ('color_v', [v, {}]),
        ]
    elif version <= 1 and cname == 'change_radius':
        if rawvalue == '0.0':
            return []
        raise Obsolete('change_radius is not supported any more')
    elif version <= 2 and cname == 'adapt_color_from_image':
        if rawvalue == '0.0':
            return []
        raise Obsolete(
            'adapt_color_from_image is obsolete, ignored;'
            ' use smudge and smudge_length instead'
        )
    elif version <= 1 and cname == 'painting_time':
        return []

    if version <= 1 and cname == 'speed':
        cname = 'speed1'
    parts = rawvalue.split('|')
    basevalue = float(parts[0])
    input_points = {}
    for part in parts[1:]:
        inputname, rawpoints = part.strip().split(' ', 1)
        if version <= 1:
            points = _oldfmt_parse_points_v1(rawpoints)
        else:
            points = _oldfmt_parse_points_v2(rawpoints)
        assert len(points) >= 2
        input_points[inputname] = points
    return [(cname, [float(basevalue), input_points])]


def _oldfmt_parse_points_v1(rawpoints):
    """Parses the points list format from v1"""
    points_seq = [float(f) for f in rawpoints.split()]
    points = [(0, 0)]
    while points_seq:
        x = points_seq.pop(0)
        y = points_seq.pop(0)
        if x == 0:
            break
        assert x > points[-1][0]
        points.append((x, y))
    return points


def _oldfmt_parse_points_v2(rawpoints):
    """Parses the newer points list format of v2 and beyond."""
    points = []
    for s in rawpoints.split(', '):
        s = s.strip()
        if not (s.startswith('(') and s.endswith(')') and ' ' in s):
            return '(x y) expected, got "%s"' % s
        s = s[1:-1]
        x, y = [float(ss) for ss in s.split(' ')]
        points.append((x, y))
    return points


def _oldfmt_transform_y(valuepair, func):
    """Used during migration from earlier versions."""
    basevalue, input_points = valuepair
    basevalue = func(basevalue)
    input_points_new = {}
    for inputname, points in input_points.items():
        points_new = [(x, func(y)) for x, y in points]
        input_points_new[inputname] = points_new
    return [basevalue, input_points_new]


# Class defs:

class BrushInfo (object):
    """Fully parsed description of a brush.
    """

    def __init__(self, string=None, default_overrides=None):
        """Construct a BrushInfo object, optionally parsing it.

        :param string: optional json string to load info from
        :param default_overrides: optional dict of
        "canonical setting name -> (BrushSettingInfo -> value)" mappings,
        each used to change the default values of a settings.
        """
        super(BrushInfo, self).__init__()
        self.settings = {}
        self.undefined_settings = set()
        self.cache_str = None
        self.observers = []
        self.default_overrides = default_overrides
        for s in brushsettings.settings:
            self.reset_setting(s.cname)
        self.observers.append(self.settings_changed_cb)
        self.observers_hidden = []
        self.pending_updates = set()
        if string:
            self.load_from_string(string)

    def settings_changed_cb(self, settings):
        self.cache_str = None

    def clone(self):
        """Returns a deep-copied duplicate."""
        res = BrushInfo()
        res.load_from_brushinfo(self)
        return res

    def load_from_brushinfo(self, other):
        """Updates the brush's Settings from (a clone of) ``brushinfo``."""
        self.settings = copy.deepcopy(other.settings)
        self.default_overrides = other.default_overrides
        self.undefined_settings = set(other.undefined_settings)
        for f in self.observers:
            f(ALL_SETTINGS)
        self.cache_str = other.cache_str

    def load_defaults(self):
        """Load default brush settings, dropping all current settings."""
        self.begin_atomic()
        self.settings = {}
        for s in brushsettings.settings:
            self.reset_setting(s.cname)
        self.end_atomic()

    def reset_setting(self, cname):
        s = brushsettings.settings_dict[cname]
        if self.default_overrides and cname in self.default_overrides:
            override = self.default_overrides[cname]
            basevalue = override(s)
        else:
            basevalue = s.default

        if cname == 'opaque_multiply':
            # make opaque depend on pressure by default
            input_points = {'pressure': [(0.0, 0.0), (1.0, 1.0)]}
        else:
            input_points = {}
        self.settings[cname] = [basevalue, input_points]
        for f in self.observers:
            f(set([cname]))

    def reset_if_undefined(self, cname):
        if cname in self.undefined_settings:
            self.reset_setting(cname)

    def to_json(self):
        settings = dict(self.settings)

        # Fields we save that aren't really brush engine settings
        parent_brush_name = settings.pop('parent_brush_name', '')
        brush_group = settings.pop('group', '')
        description = settings.pop('description', '')
        notes = settings.pop('notes', '')

        # The comment we save is always the same
        settings.pop('comment', '')

        # Make the contents of each setting a bit more explicit
        for k, v in list(settings.items()):
            base_value, inputs = v
            settings[k] = {'base_value': base_value, 'inputs': inputs}

        document = {
            'version': 3,
            'comment': """MyPaint brush file""",
            'parent_brush_name': parent_brush_name,
            'settings': settings,
            'group': brush_group,
            'notes': notes,
            'description': description,
        }
        return json.dumps(document, sort_keys=True, indent=4)

    def from_json(self, json_string):
        """Loads settings from a JSON string.

        >>> from glob import glob
        >>> for p in glob("tests/brushes/v3/*.myb"):
        ...     with open(p, "rb") as fp:
        ...         bstr = fp.read()
        ...         ustr = bstr.decode("utf-8")
        ...     b1 = BrushInfo()
        ...     b1.from_json(bstr)
        ...     b1 = BrushInfo()
        ...     b1.from_json(ustr)

        See also load_from_string(), which can handle the old v2 format.

        Accepts both unicode and byte strings. Byte strings are assumed
        to be encoded as UTF-8 when any decoding's needed.

        """

        # Py3: Ubuntu Trusty's 3.4.3 json.loads() requires unicode strs.
        # Layer Py3, and Py2 is OK with either.
        if not isinstance(json_string, unicode):
            if not isinstance(json_string, bytes):
                raise ValueError("Need either a str or a bytes object")
            json_string = json_string.decode("utf-8")

        brush_def = json.loads(json_string)
        if brush_def.get('version', 0) < 3:
            raise BrushInfo.ParseError(
                'brush is not compatible with this version of mypaint '
                '(json file version=%r)' % (brush_def.get('version'),)
            )

        # settings not in json_string must still be present in self.settings
        self.load_defaults()

        # settings not defined in the json
        self.undefined_settings = BRUSH_SETTINGS.difference(
            set(brush_def['settings'].keys())
        )
        # MyPaint expects that each setting has an array, where
        # index 0 is base value, and index 1 is inputs
        for k, v in brush_def['settings'].items():
            base_value, inputs = v['base_value'], v['inputs']
            if k not in self.settings:
                logger.warning('ignoring unknown brush setting %r', k)
                continue
            self.settings[k] = [base_value, inputs]

        # Non-libmypaint string fields
        for cname in STRING_VALUE_SETTINGS:
            self.settings[cname] = brush_def.get(cname, '')
        # FIXME: Who uses "group"?
        # FIXME: Brush groups are stored externally in order.conf,
        # FIXME: is that one redundant?

    @staticmethod
    def brush_string_inverted_eotf(brush_string):
        if isinstance(brush_string, bytes):
            brush_string = brush_string.decode("utf-8")
        try:
            brush = json.loads(brush_string)
            bsett = brush['settings']
            k = 'base_value'
            hsv = bsett['color_h'][k], bsett['color_s'][k], bsett['color_v'][k]
            h, s, v = helpers.transform_hsv(hsv, 1.0 / 2.2)
            bsett['color_h'][k] = h
            bsett['color_s'][k] = s
            bsett['color_v'][k] = v
            return json.dumps(brush)
        except Exception:
            logger.exception("Failed to invert color in brush string")
            return brush_string

    def load_from_string(self, settings_str):
        """Load a setting string, overwriting all current settings."""

        settings_unicode = settings_str
        if not isinstance(settings_unicode, unicode):
            if not isinstance(settings_unicode, bytes):
                raise ValueError("Need either a str or a bytes object")
            settings_unicode = settings_unicode.decode("utf-8")

        if settings_unicode.startswith(u'{'):
            # new json-based brush format
            self.from_json(settings_str)
        elif settings_unicode.startswith(u'#'):
            # old brush format
            self._load_old_format(settings_str)
        else:
            raise BrushInfo.ParseError('brush format not recognized')

        for f in self.observers:
            f(ALL_SETTINGS)
        self.cache_str = settings_str

    def _load_old_format(self, settings_str):
        """Loads brush settings in the old (v2) format.

        >>> from glob import glob
        >>> for p in glob("tests/brushes/v2/*.myb"):
        ...     with open(p, "rb") as fp:
        ...         bstr = fp.read()
        ...         ustr = bstr.decode("utf-8")
        ...     b1 = BrushInfo()
        ...     b1._load_old_format(bstr)
        ...     b2 = BrushInfo()
        ...     b2._load_old_format(ustr)

        Accepts both unicode and byte strings. Byte strings are assumed
        to be encoded as UTF-8 when any decoding's needed.

        """

        # Py2 is happy natively comparing unicode with str, no encode
        # needed. For Py3, need to parse as str so that updated dict
        # keys can be compared sensibly with stuff written by other
        # code.

        if not isinstance(settings_str, unicode):
            if not isinstance(settings_str, bytes):
                raise ValueError("Need either a str or a bytes object")
            if PY3:
                settings_str = settings_str.decode("utf-8")

        # Split out the raw settings and grab the version we're dealing with
        rawsettings = []
        errors = []
        version = 1  # for files without a 'version' field
        for line in settings_str.split('\n'):
            try:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                cname, rawvalue = line.split(' ', 1)
                if cname == 'version':
                    version = int(rawvalue)
                    if version > OLDFORMAT_BRUSHFILE_VERSION:
                        raise BrushInfo.ParseError(
                            "This brush is not in the old format "
                            "supported (version > {})".format(
                                OLDFORMAT_BRUSHFILE_VERSION,
                            )
                        )
                else:
                    rawsettings.append((cname, rawvalue))
            except Exception as e:
                errors.append((line, str(e)))

        # Parse each pair
        self.load_defaults()
        # compatibility hack: keep disabled for old brushes,
        # but still use non-zero default
        self.settings['anti_aliasing'][0] = 0.0
        num_parsed = 0
        settings_loaded = set()
        for rawcname, rawvalue in rawsettings:
            try:
                cnamevaluepairs = _oldfmt_parse_value(
                    rawvalue,
                    rawcname,
                    version,
                )
                num_parsed += 1
                for cname, value in cnamevaluepairs:
                    if cname in brushsettings.settings_migrate:
                        cname, func = brushsettings.settings_migrate[cname]
                        if func:
                            value = _oldfmt_transform_y(value, func)
                    self.settings[cname] = value
                    settings_loaded.add(cname)
            except Exception as e:
                line = "%s %s" % (rawcname, rawvalue)
                errors.append((line, str(e)))
        if errors:
            for error in errors:
                logger.warning(error)
        if num_parsed == 0:
            raise BrushInfo.ParseError(
                "old brush file format parser did not find "
                "any brush settings in this file",
            )
        self.undefined_settings = BRUSH_SETTINGS.difference(settings_loaded)

    def save_to_string(self):
        """Serialise brush information to a string. Result is cached."""
        if self.cache_str:
            return self.cache_str

        res = self.to_json()

        self.cache_str = res
        return res

    def get_base_value(self, cname):
        return self.settings[cname][0]

    def get_points(self, cname, input, readonly=False):
        res = self.settings[cname][1].get(input, ())
        if not readonly:  # slow
            res = copy.deepcopy(res)
        return res

    def set_base_value(self, cname, value):
        assert cname in BRUSH_SETTINGS
        assert not math.isnan(value)
        assert not math.isinf(value)
        if self.settings[cname][0] != value:
            if cname in self.undefined_settings:
                self.undefined_settings.remove(cname)
            self.settings[cname][0] = value
            for f in self.observers:
                f(set([cname]))

    def set_points(self, cname, input, points):
        assert cname in BRUSH_SETTINGS
        if cname in self.undefined_settings:
            self.undefined_settings.remove(cname)
        points = tuple(points)
        d = self.settings[cname][1]
        if points:
            d[input] = copy.deepcopy(points)
        elif input in d:
            d.pop(input)

        for f in self.observers:
            f(set([cname]))

    def set_setting(self, cname, value):
        self.settings[cname] = copy.deepcopy(value)
        if cname in self.undefined_settings:
            self.undefined_settings.remove(cname)
        for f in self.observers:
            f(set([cname]))

    def get_setting(self, cname):
        return copy.deepcopy(self.settings[cname])

    def get_string_property(self, name):
        value = self.settings.get(name, None)
        if value is None:
            return None
        return unicode(value)

    def set_string_property(self, name, value):
        assert name in STRING_VALUE_SETTINGS
        if value is None:
            self.settings.pop(name, None)
        else:
            assert isinstance(value, str) or isinstance(value, unicode)
            self.settings[name] = unicode(value)
        for f in self.observers:
            f(set([name]))

    def has_only_base_value(self, cname):
        """Return whether a setting is constant for this brush."""
        for i in brushsettings.inputs:
            if self.has_input(cname, i.name):
                return False
        return True

    def has_large_base_value(self, cname, threshold=0.9):
        return self.get_base_value(cname) > threshold

    def has_small_base_value(self, cname, threshold=0.1):
        return self.get_base_value(cname) < threshold

    def has_input(self, cname, input):
        """Return whether a given input is used by some setting."""
        points = self.get_points(cname, input, readonly=True)
        return bool(points)

    def begin_atomic(self):
        self.observers_hidden.append(self.observers[:])
        del self.observers[:]
        self.observers.append(self.add_pending_update)

    def add_pending_update(self, settings):
        self.pending_updates.update(settings)

    def end_atomic(self):
        self.observers[:] = self.observers_hidden.pop()
        pending = self.pending_updates.copy()
        if pending:
            self.pending_updates.clear()
            for f in self.observers:
                f(pending)

    def get_color_hsv(self):
        h = self.get_base_value('color_h')
        s = self.get_base_value('color_s')
        v = self.get_base_value('color_v')
        assert not math.isnan(h)
        return h, s, v

    def set_color_hsv(self, hsv):
        if not hsv:
            return
        self.begin_atomic()
        try:
            h, s, v = hsv
            self.set_base_value('color_h', h)
            self.set_base_value('color_s', s)
            self.set_base_value('color_v', v)
        finally:
            self.end_atomic()

    def set_color_rgb(self, rgb):
        self.set_color_hsv(helpers.rgb_to_hsv(*rgb))

    def get_color_rgb(self):
        hsv = self.get_color_hsv()
        return helpers.hsv_to_rgb(*hsv)

    def is_eraser(self):
        return self.has_large_base_value("eraser")

    def is_alpha_locked(self):
        return self.has_large_base_value("lock_alpha")

    def is_colorize(self):
        return self.has_large_base_value("colorize")

    def matches(self, other, ignore=_BRUSHINFO_MATCH_IGNORES):
        s1 = self.settings.copy()
        s2 = other.settings.copy()
        for k in ignore:
            s1.pop(k, None)
            s2.pop(k, None)
        return s1 == s2


class Brush (mypaintlib.PythonBrush):
    """A brush, capable of painting to a surface

    Low-level extension of the C++ brush class, propagating all changes
    made to a BrushInfo instance down into the C brush struct.

    """

    HSV_CNAMES = ('color_h', 'color_s', 'color_v')
    HSV_SET = set(HSV_CNAMES)

    def __init__(self, brushinfo):
        super(Brush, self).__init__()
        self.brushinfo = brushinfo
        brushinfo.observers.append(self._update_from_brushinfo)
        self._update_from_brushinfo(ALL_SETTINGS)

    def stroke_to(self, *args):
        """ Delegates to mypaintlib with information about color space

        Checks whether color transforms should be done in linear sRGB
        so that HSV/HSL adjustments can be handled correctly.
        """
        if eotf() == 1.0:
            return super(Brush, self).stroke_to(*args)
        else:
            return super(Brush, self).stroke_to_linear(*args)

    def _update_from_brushinfo(self, settings):
        """Updates changed low-level settings from the BrushInfo"""

        # When eotf != 1.0, store transformed hsv values in the backend.
        transform = eotf() != 1.0
        if transform and any(hsv in settings for hsv in self.HSV_CNAMES):
            self._transform_brush_color()
            # Clear affected settings so the transformation
            # is not undone in the next step.
            # Note: x = x - y is not equivalent to x -= y here.
            settings = settings - self.HSV_SET

        for cname in settings:
            self._update_setting_from_brushinfo(cname)

    def _transform_brush_color(self):
        """ Apply eotf transform to the backend color.

        By only applying the transform here, the issue of
        strokemap and brush color consistency between new
        and old color rendering modes does not arise.
        """
        hsv_orig = (self.brushinfo.get_base_value(k) for k in self.HSV_CNAMES)
        h, s, v = helpers.transform_hsv(hsv_orig, eotf())
        settings_dict = brushsettings.settings_dict
        self.set_base_value(settings_dict['color_h'].index, h)
        self.set_base_value(settings_dict['color_s'].index, s)
        self.set_base_value(settings_dict['color_v'].index, v)

    def _update_setting_from_brushinfo(self, cname):
        setting = brushsettings.settings_dict.get(cname)
        if not setting:
            return
        base = self.brushinfo.get_base_value(cname)
        self.set_base_value(setting.index, base)
        for input in brushsettings.inputs:
            points = self.brushinfo.get_points(cname, input.name,
                                               readonly=True)
            assert len(points) != 1
            self.set_mapping_n(setting.index, input.index, len(points))
            for i, (x, y) in enumerate(points):
                self.set_mapping_point(setting.index, input.index, i, x, y)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# This file is part of MyPaint.
# Copyright (C) 2011-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Layout constants and helper functions for common widgets."""

from __future__ import division, print_function
import functools

from lib.gibindings import Gtk
from lib.gibindings import Gdk


# Exact icon sizes

ICON_SIZE_LARGE = Gtk.IconSize.LARGE_TOOLBAR   # 24x24, the docs promise
ICON_SIZE_SMALL = Gtk.IconSize.SMALL_TOOLBAR   # 16x16


# Spacing constants

SPACING_CRAMPED = 3    # for use in dockables only
SPACING_TIGHT = 2 * SPACING_CRAMPED
SPACING = 2 * SPACING_TIGHT
SPACING_LOOSE = 3*SPACING_TIGHT


def borderless_button(stock_id=None, icon_name=None, size=ICON_SIZE_SMALL,
                      tooltip=None, action=None):
    """Create a button styled to be borderless.

    >>> borderless_button(icon_name="mypaint")  # doctest: +ELLIPSIS
    <Gtk.Button...>

    """
    button = Gtk.Button()
    if stock_id is not None:
        image = Gtk.Image()
        image.set_from_stock(stock_id, size)
        set_margins(image, 0)
        button.add(image)
    elif icon_name is not None:
        image = Gtk.Image()
        image.set_from_icon_name(icon_name, size)
        set_margins(image, 0)
        button.add(image)
    elif action is not None:
        button.set_related_action(action)
        if button.get_child() is not None:
            button.remove(button.get_child())
        img = action.create_icon(size)
        img.set_padding(4, 4)
        set_margins(img, 0)
        button.add(img)
    button.set_relief(Gtk.ReliefStyle.NONE)
    button.set_can_default(False)
    button.set_can_focus(False)
    set_margins(button, 0)
    if tooltip is not None:
        button.set_tooltip_text(tooltip)
    elif action is not None:
        button.set_tooltip_text(action.get_tooltip())
    cssprov = Gtk.CssProvider()
    cssprov.load_from_data(b"GtkButton { padding: 0px; margin: 0px; }")
    style = button.get_style_context()
    style.add_provider(cssprov, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    button.set_has_tooltip(tooltip is not None)
    return button


def set_margins(widget, all_=0, tb=None, lr=None,
                t=None, b=None, l=None, r=None):  # noqa: E741
    """Set margins compatibly on a widget.

    >>> w = Gtk.Label(label="i have wide margins")
    >>> set_margins(w, 42)

    Works around Gtk's deprecation of gtk_widget_set_margin_{left,right}
    in version 3.12.

    """
    top = bot = left = right = 0
    if all_ is not None:
        top = bot = left = right = int(all_)
    if tb is not None:
        top = bot = int(tb)
    if lr is not None:
        left = right = int(lr)
    if t is not None:
        top = int(t)
    if b is not None:
        bot = int(b)
    if l is not None:
        left = int(l)
    if r is not None:
        right = int(r)
    try:
        widget.set_margin_start(left)
        widget.set_margin_end(right)
    except AttributeError:
        widget.set_margin_left(left)
        widget.set_margin_right(right)
    widget.set_margin_top(top)
    widget.set_margin_bottom(bot)


def inline_toolbar(app, tool_defs):
    """Builds a styled inline toolbar"""
    bar = Gtk.Toolbar(
        show_arrow = False,
        icon_size = ICON_SIZE_SMALL,
    )
    bar.set_style(Gtk.ToolbarStyle.ICONS)
    styles = bar.get_style_context()
    styles.add_class(Gtk.STYLE_CLASS_INLINE_TOOLBAR)
    for action_name, override_icon in tool_defs:
        action = app.find_action(action_name)
        toolitem = Gtk.ToolButton()
        toolitem.set_related_action(action)
        if override_icon:
            toolitem.set_icon_name(override_icon)
        bar.insert(toolitem, -1)
        bar.child_set_property(toolitem, "expand", True)
        bar.child_set_property(toolitem, "homogeneous", True)
    bar.set_vexpand(False)
    bar.set_hexpand(True)
    return bar


class MenuButtonToolItem (Gtk.ToolItem):
    """ToolItem which contains a Gtk.MenuButton"""

    def __init__(self):
        Gtk.ToolItem.__init__(self)
        menubtn = Gtk.MenuButton()
        self.add(menubtn)
        self.connect("realize", self._realize_cb)
        menubtn.set_always_show_image(True)
        menubtn.set_relief(Gtk.ReliefStyle.NONE)
        self._menubutton = menubtn
        self.menu = None  #: Populate with the menu to show before realize

    def _realize_cb(self, widget):
        action = self.get_related_action()
        icon_name = action.get_icon_name()
        image = Gtk.Image.new_from_icon_name(
            icon_name,
            get_toolbar_icon_size(),
        )
        self._menubutton.set_image(image)
        if self.menu:
            self._menubutton.set_popup(self.menu)


def get_toolbar_icon_size():
    from gui.application import get_app
    app = get_app()
    size = str(app.preferences.get("ui.toolbar_icon_size", "large"))
    if size.lower() == 'small':
        return ICON_SIZE_SMALL
    else:
        return ICON_SIZE_LARGE


def with_wait_cursor(func):
    """python decorator that adds a wait cursor around a function"""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        wait_cursor = Gdk.Cursor.new(Gdk.CursorType.WATCH)
        toplevels = Gtk.Window.list_toplevels()
        toplevels = [t for t in toplevels if t.get_window() is not None]
        for toplevel in toplevels:
            toplevel_win = toplevel.get_window()
            if toplevel_win is not None:
                toplevel_win.set_cursor(wait_cursor)
            toplevel.set_sensitive(False)
        try:
            return func(self, *args, **kwargs)
            # gtk main loop may be called in here...
        finally:
            for toplevel in toplevels:
                toplevel.set_sensitive(True)
                # ... which is why we need this check
                toplevel_win = toplevel.get_window()
                if toplevel_win is not None:
                    toplevel_win.set_cursor(None)
    return wrapper



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2009-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Brush selection/organizer windows

These can be used for selecting brushes, and can be docked into the sidebar.
They are responsible for ordering, loading and saving brush lists.

"""

## Imports

from __future__ import division, print_function
import logging

from lib.gibindings import Gtk
from lib.gibindings import GdkPixbuf
from lib.gibindings import GLib

from lib.gettext import C_
from lib.gettext import ngettext

from . import pixbuflist
from . import dialogs
from . import brushmanager
from .toolstack import SizedVBoxToolWidget
from . import widgets

logger = logging.getLogger(__name__)


## Helper functions

def managedbrush_idfunc(managedbrush):
    """Returns the id of a ManagedBrush."""
    return managedbrush.name


def managedbrush_namefunc(managedbrush):
    """Returns tooltip of a ManagedBrush."""
    template = "{name}"
    if managedbrush.description:
        template = "{name}\n{description}"
    return template.format(
        name = managedbrush.get_display_name(),
        description = managedbrush.description,
    )


def managedbrush_pixbuffunc(managedbrush):
    """Returns pixbuf preview of a ManagedBrush."""
    return managedbrush.preview


## Class definitions


class BrushList (pixbuflist.PixbufList):
    """Flowed grid of brush icons showing a group, click to set the brush"""

    ICON_SIZE = 48
    MIN_WIDTH_NICONS = 1
    NATURAL_WIDTH_NICONS = 4
    MIN_HEIGHT_NICONS = 1

    def __init__(self, app, group):
        """Construct, for a named group."""
        self.app = app
        self.bm = app.brushmanager
        self.group = group
        s = self.ICON_SIZE
        super(BrushList, self).__init__(
            self.brushes, s, s,
            namefunc=managedbrush_namefunc,
            pixbuffunc=managedbrush_pixbuffunc,
            idfunc=managedbrush_idfunc,
        )
        self.set_selected(self.bm.selected_brush)
        self.bm.groups_changed += self._groups_changed_cb
        self.bm.brushes_changed += self._brushes_changed_cb
        self.bm.brush_selected += self._brush_selected_cb
        self.item_selected += self._item_selected_cb
        self.item_popup += self._item_popup_cb

    @property
    def brushes(self):
        """The list of brushes being shown.

        The returned list belongs to the main app's BrushManager,
        and is created on demand if there is no such group.
        If you reorder it or remove brushes,
        you must call the BrushManager's brushes_changed() method too.

        """
        group_name = self.group
        return self.bm.get_group_brushes(group_name)

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH

    def do_get_preferred_width(self):
        return (self.MIN_WIDTH_NICONS * self.ICON_SIZE,
                self.NATURAL_WIDTH_NICONS * self.ICON_SIZE)

    def do_get_preferred_height_for_width(self, width):
        icons_wide = max(1, int(width // self.ICON_SIZE))
        num_brushes = len(self.brushes)
        icons_tall = max(int(num_brushes // icons_wide),
                         max(self.MIN_HEIGHT_NICONS, 1))
        if icons_tall * icons_wide < num_brushes:
            icons_tall += 1
        return (icons_tall * self.ICON_SIZE,
                icons_tall * self.ICON_SIZE)

    def _groups_changed_cb(self, bm):
        # In case the group has been deleted and recreated, we do this
        group_name = self.group
        group_brushes = self.bm.groups.get(group_name, [])
        self.itemlist = group_brushes
        self.update()
        # See https://github.com/mypaint/mypaint/issues/654

    def _brushes_changed_cb(self, bm, brushes):
        # CARE: this might be called in response to the group being deleted.
        # Don't recreate it by accident.
        group_name = self.group
        group_brushes = self.bm.groups.get(group_name)
        if brushes is group_brushes:
            self.update()

    def _brush_selected_cb(self, bm, managed_brush, brushinfo):
        self.set_selected(managed_brush)

    def remove_brush(self, brush):
        self.brushes.remove(brush)
        self.bm.brushes_changed(self.brushes)

    def insert_brush(self, idx, brush):
        self.brushes.insert(idx, brush)
        self.bm.brushes_changed(self.brushes)

    def button_press_cb(self, widget, event):
        device = event.get_source_device()
        self.app.device_monitor.device_used(device)
        return super(BrushList, self).button_press_cb(widget, event)

    def drag_begin_cb(self, widget, context):
        preview = self.bm.selected_brush.preview
        preview = preview.scale_simple(
            preview.get_width() // 2,
            preview.get_height() // 2,
            GdkPixbuf.InterpType.BILINEAR,
        )
        Gtk.drag_set_icon_pixbuf(context, preview, 0, 0)
        super(BrushList, self).drag_begin_cb(widget, context)

    def on_drag_data(self, copy, source_widget, brush_dragid, target_idx):
        assert source_widget, 'cannot handle drag data from another app'
        brush = None
        for b in source_widget.brushes:
            b_dragid = source_widget.idfunc(b)
            if b_dragid == brush_dragid:
                brush = b
                break
        if brush is None:
            logger.error(
                "No brush with dragid=%r in drag source widget",
                brush_dragid,
            )
            return False
        if source_widget is self:
            copy = False
        elif brush in self.brushes:
            return True
        if not copy:
            source_widget.remove_brush(brush)
        self.insert_brush(target_idx, brush)
        return True

    def _item_selected_cb(self, self_, brush):
        # brush changed on harddisk?
        if brush.reload_if_changed():
            for brushes in self.bm.groups.values():
                self.bm.brushes_changed(brushes)
        self.bm.select_brush(brush)

    def _item_popup_cb(self, self_, brush):
        time = Gtk.get_current_event_time()
        menu = BrushPopupMenu(self, brush)
        menu.show_all()
        menu.popup(
            parent_menu_shell = None,
            parent_menu_item = None,
            func = None,
            button = 3,
            activate_time = time,
            data = None,
        )


class BrushPopupMenu (Gtk.Menu):
    """Pop-up menu for brush actions."""

    def __init__(self, bl, brush):
        super(BrushPopupMenu, self).__init__()
        self._brush = brush
        self._brushlist = bl
        self._app = bl.app
        faves = bl.bm.get_group_brushes(brushmanager.FAVORITES_BRUSH_GROUP)
        if brush not in faves:
            item = Gtk.MenuItem(label=C_(
                "brush group: context menu for a single brush",
                "Add to Favorites",
            ))
            item.connect("activate", self._favorite_cb)
            self.append(item)
        else:
            item = Gtk.MenuItem(label=C_(
                "brush group: context menu for a single brush",
                "Remove from Favorites",
            ))
            item.connect("activate", self._unfavorite_cb)
            self.append(item)

        if bl.group != brushmanager.FAVORITES_BRUSH_GROUP:
            item = Gtk.MenuItem(label=C_(
                "brush group: context menu for a single brush",
                "Clone",
            ))
            item.connect("activate", self._clone_cb)
            self.append(item)

        item = Gtk.MenuItem(label=C_(
            "brush group: context menu for a single brush",
            "Edit Brush Settings",
        ))
        item.connect("activate", self._edit_cb)
        self.append(item)

        if bl.group != brushmanager.FAVORITES_BRUSH_GROUP:
            item = Gtk.MenuItem(label=C_(
                "brush group: context menu for a single brush",
                "Remove from Group",
            ))
            item.connect("activate", self._remove_cb)
            self.append(item)

    def _favorite_cb(self, menuitem):
        bl = self._brushlist
        brush = self._brush
        # Update the faves group if the brush isn't already there.
        faves_group_name = brushmanager.FAVORITES_BRUSH_GROUP
        faves = bl.bm.get_group_brushes(faves_group_name)
        if brush not in faves:
            faves.append(brush)
            bl.bm.brushes_changed(faves)
            bl.bm.save_brushorder()
        # Show the faves group
        workspace = self._app.workspace
        gtype_name = BrushGroupTool.__gtype_name__
        params = (faves_group_name,)
        workspace.reveal_tool_widget(gtype_name, params)
        # Highlight the (possibly copied) brush
        bl.bm.select_brush(brush)

    def _unfavorite_cb(self, menuitem):
        bl = self._brushlist
        brush = self._brush
        faves = bl.bm.get_group_brushes(brushmanager.FAVORITES_BRUSH_GROUP)
        try:
            faves.remove(brush)
        except ValueError:
            return
        bl.bm.brushes_changed(faves)
        bl.bm.save_brushorder()

    def _clone_cb(self, menuitem):
        bl = self._brushlist
        brush = self._brush
        # Pick a nice unique name
        new_name = C_(
            "brush group: context menu: unique names for cloned brushes",
            u"{original_name} copy"
        ).format(
            original_name = brush.name,
        )
        uniquifier = 0
        while bl.bm.get_brush_by_name(new_name):
            uniquifier += 1
            new_name = C_(
                "brush group: context menu: unique names for cloned brushes",
                u"{original_name} copy {n}"
            ).format(
                name = brush.name,
                n = uniquifier,
            )
        # Make a copy and insert it near the original
        brush_copy = brush.clone(new_name)
        index = bl.brushes.index(brush) + 1
        bl.insert_brush(index, brush_copy)
        brush_copy.save()
        bl.bm.save_brushorder()
        # Select the copy, for highlighting
        bl.bm.select_brush(brush_copy)

    def _edit_cb(self, menuitem):
        bl = self._brushlist
        brush = self._brush
        bl.bm.select_brush(brush)
        brush_editor = self._app.brush_settings_window
        brush_editor.show()

    def _remove_cb(self, menuitem):
        bl = self._brushlist
        brush = self._brush
        msg = C_(
            "brush group: context menu: remove from group",
            u"Really remove brush “{brush_name}” "
            u"from group “{group_name}”?"
        ).format(
            brush_name = brush.name,
            group_name = bl.group,
        )
        if not dialogs.confirm(bl, msg):
            return
        bl.remove_brush(brush)


class BrushGroupTool (SizedVBoxToolWidget):
    """Dockable tool widget showing just one BrushGroup"""

    __gtype_name__ = "MyPaintBrushGroupTool"

    ## Construction and updating

    def __init__(self, group):
        """Construct, to show a named group"""
        SizedVBoxToolWidget.__init__(self)
        self._group = group
        self._scrolls = Gtk.ScrolledWindow()
        self._dialog = None
        self._brush_list = None
        from gui.application import get_app
        self._app = get_app()
        if group not in self._app.brushmanager.groups:
            raise ValueError("No group named %r" % group)
        self.pack_start(self._scrolls, True, True, 0)
        self._update_brush_list()

    def _update_brush_list(self):
        """Updates the brush list to match the group name"""
        if self._brush_list:
            self._brush_list.destroy()
        viewport = self._scrolls.get_child()
        if viewport:
            self._scrolls.remove(viewport)
            viewport.destroy()
        self._brush_list = BrushList(self._app, self._group)
        self._scrolls.add(self._brush_list)
        self._brush_list.show_all()

    ## Tool widget properties and methods

    @property
    def tool_widget_title(self):
        return brushmanager.translate_group_name(self._group)

    @property
    def tool_widget_description(self):
        if self._group not in self._app.brushmanager.groups:
            return None
        nbrushes = len(self._app.brushmanager.groups[self._group])
        # TRANSLATORS: number of brushes in a brush group, for tooltips
        return ngettext("%d brush", "%d brushes", nbrushes) % (nbrushes,)

    @property
    def tool_widget_icon_name(self):
        return "mypaint-brushes-symbolic"  # fallback only

    def tool_widget_get_icon_pixbuf(self, size):
        if self._group not in self._app.brushmanager.groups:
            return None
        brushes = self._app.brushmanager.groups[self._group]
        if not brushes:
            return None
        icon = brushes[0].preview
        if size == icon.get_width() and size == icon.get_height():
            return icon.copy()
        else:
            return icon.scale_simple(size, size, GdkPixbuf.InterpType.BILINEAR)

    def tool_widget_properties(self):
        """Run the properties dialog"""
        if not self._dialog:
            title = C_(
                "brush group properties dialog: title",
                # TRANSLATORS: properties dialog for the current brush group
                u"Group \u201C{group_name}\u201D",
            ).format(
                group_name = self._group,
            )
            dia = Gtk.Dialog(
                title=title,
                modal=True,
                destroy_with_parent=True,
                window_position=Gtk.WindowPosition.MOUSE,
            )
            dia.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT)
            btn = Gtk.Button(label=C_(
                "brush group properties dialog: action buttons",
                "Rename Group",
            ))
            btn.connect("clicked", self._rename_cb)
            dia.vbox.pack_start(btn, False, False, 0)
            btn = Gtk.Button(label=C_(
                "brush group properties dialog: action buttons",
                "Export as Zipped Brushset",
            ))
            btn.connect("clicked", self._export_cb)
            dia.vbox.pack_start(btn, False, False, 0)
            btn = Gtk.Button(label=C_(
                "brush group properties dialog: action buttons",
                "Delete Group",
            ))
            btn.connect("clicked", self._delete_cb)
            dia.vbox.pack_start(btn, False, False, 0)
            dia.vbox.show_all()
            self._dialog = dia
        self._dialog.set_transient_for(self.get_toplevel())
        self._dialog.run()
        self._dialog.hide()

    ## Properties dialog action callbacks

    def _rename_cb(self, widget):
        """Properties dialog rename callback"""
        # XXX Because of the way this works, groups can only be renamed from
        # XXX    the widget's properties dialog at present. Maybe that's OK.
        self._dialog.hide()
        old_group = self._group
        new_group = dialogs.ask_for_name(
            self,
            C_("brush group rename dialog: title", "Rename Group"),
            old_group,
        )
        if not new_group:
            return
        if old_group not in self._app.brushmanager.groups:
            return
        if new_group not in self._app.brushmanager.groups:
            self._app.brushmanager.rename_group(old_group, new_group)
            self._group = new_group
            workspace = self._app.workspace
            gtype_name = self.__gtype_name__
            workspace.update_tool_widget_params(gtype_name,
                                                (old_group,), (new_group,))
            self._update_brush_list()
        else:
            dialogs.error(self, C_(
                'brush group rename',
                'A group with this name already exists!',
            ))

    def _delete_cb(self, widget):
        """Properties dialog delete callback"""
        self._dialog.hide()
        name = brushmanager.translate_group_name(self._group)
        msg = C_(
            "brush group delete",
            u"Really delete group \u201C{group_name}\u201D?",
        ).format(
            group_name = name,
        )
        bm = self._app.brushmanager
        if not dialogs.confirm(self, msg):
            return
        bm.delete_group(self._group)
        if self._group not in bm.groups:
            GLib.idle_add(
                self._remove_panel_idle_cb,
                self.__gtype_name__, (self._group,),
            )
            return
        # Special groups like "Deleted" cannot be deleted,
        # but the error message is very confusing in that case...
        msg = C_(
            "brush group delete",
            u"Could not delete group \u201C{group_name}\u201D.\n"
            u"Some special groups cannot be deleted.",
        ).format(
            group_name = name,
        )
        dialogs.error(self, msg)

    def _remove_panel_idle_cb(self, typespec, paramspec):
        self._app.workspace.remove_tool_widget(typespec, paramspec)
        return False

    def _export_cb(self, widget):
        """Properties dialog export callback"""
        self._dialog.hide()
        format_id, filename = dialogs.save_dialog(
            C_("brush group export dialog: title", "Export Brushes"),
            None,
            [(
                C_(
                    "brush group export dialog",
                    "MyPaint brush package (*.zip)",
                ),
                "*.zip",
            )],
            default_format=(0, ".zip"),
        )
        if filename is not None:
            self._app.brushmanager.export_group(self._group, filename)


class BrushGroupsMenu (Gtk.Menu):
    """Dynamic menu containing all the brush groups"""

    def __init__(self):
        super(BrushGroupsMenu, self).__init__()
        from gui.application import get_app
        self.app = get_app()
        # Static items
        item = Gtk.SeparatorMenuItem()
        self.append(item)
        item = Gtk.MenuItem(label=C_("brush groups menu", u"New Group…"))
        item.connect("activate", self._new_brush_group_cb)
        self.append(item)
        item = Gtk.MenuItem(label=C_("brush groups menu", u"Import Brushes…"))
        item.connect("activate", self.app.drawWindow.import_brush_pack_cb)
        self.append(item)
        item = Gtk.MenuItem(label=C_(
            "brush groups menu", u"Get More Brushes…")
        )
        item.connect("activate", self.app.drawWindow.download_brush_pack_cb)
        self.append(item)
        # Dynamic items
        bm = self.app.brushmanager
        self._items = {}
        self._update(bm)
        bm.groups_changed += self._update

    def _new_brush_group_cb(self, widget):
        # XXX should be moved somewhere more sensible than this
        toplevel = self.app.drawWindow
        name = dialogs.ask_for_name(
            toplevel,
            C_("new brush group dialog: title", 'Create Group'),
            '',
        )
        if name is None:
            return
        name = name.strip()
        if name:
            bm = self.app.brushmanager
            bm.create_group(name)

    def _update(self, bm):
        """Update dynamic items in response to the groups list changing"""
        for item in self._items.values():
            if item not in self:
                continue
            self.remove(item)
        activate_cb = self._brush_group_item_activate_cb
        for name in reversed(sorted(bm.groups)):
            if name in self._items:
                item = self._items[name]
            else:
                item = Gtk.ImageMenuItem()
                label = brushmanager.translate_group_name(name)
                item.set_label(label)
                item.connect("activate", activate_cb, name)
                self._items[name] = item
            self.prepend(item)
        for name, item in list(self._items.items()):
            if item in self:
                continue
            self._items.pop(name)
        self.show_all()

    def _brush_group_item_activate_cb(self, menuitem, group_name):
        workspace = self.app.workspace
        gtype_name = BrushGroupTool.__gtype_name__
        params = (group_name,)
        workspace.add_tool_widget(gtype_name, params)


class BrushGroupsToolItem (widgets.MenuButtonToolItem):
    """Toolbar item showing a dynamic dropdown BrushGroupsMenu

    This is instantiated by the app's UIManager using a FactoryAction which
    must be named "BrushGroups" (see factoryaction.py).
    """

    __gtype_name__ = "MyPaintBrushGroupsToolItem"

    def __init__(self):
        super(BrushGroupsToolItem, self).__init__()
        self.menu = BrushGroupsMenu()


class BrushGroupsMenuItem (Gtk.MenuItem):
    """Brush list menu item with a dynamic BrushGroupsMenu as its submenu

    This is instantiated by the app's UIManager using a FactoryAction which
    must be named "BrushGroups" (see factoryaction.py).
    """

    __gtype_name__ = "MyPaintBrushGroupsMenuItem"

    def __init__(self):
        super(BrushGroupsMenuItem, self).__init__()
        self._submenu = BrushGroupsMenu()
        self.set_submenu(self._submenu)
        self._submenu.show_all()



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2017-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Bulk management of layer visibility: backend and model.

The use case is a user preparing an artwork with multiple variations,
possibly with a lot shared between them.  For example, they may want one
file to store different expressions and poses for a character,
or different times of day and lighting for a landscape.

A user needing this should be able to store the set of layers they see
right now in the form of a viewing context, and assign a name to to the
context. They should be able to switch between these different contexts
at will. On each such transition, the set of layers which are visible in
the document changes accordingly. These transitions are undoable.

"""

# Imports:

from __future__ import print_function, division

import logging

from lib.gettext import C_
import lib.naming
from lib.observable import event
from lib.command import Command
from lib.pycompat import unicode


# Module vars:

logger = logging.getLogger(__name__)

NEW_VIEW_IDENT = C_(
    "layer visibility sets: default name for a user-managed view",
    u"View",
)

UNSAVED_VIEW_DISPLAY_NAME = C_(
    "layer visibility sets: text shown when no user-managed view is active",
    u"No active view",
)


# Data model classes:

class _View (object):
    """Lightweight representation of a layer viewing context.

    Views are represented as essentially just a tag.  They intentionally
    do not contain references to their layers to avoid circular refs.
    Instead, layers contain refs to their views in a private set that
    the LayerViewManager knows about.

    """

    def __init__(self, name, locked=True, **kwargs):
        super(_View, self).__init__()
        self._name = unicode(name)
        self._locked = bool(locked)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = unicode(value)

    @property
    def locked(self):
        return self._locked

    @locked.setter
    def locked(self, state):
        self._locked = bool(state)

    def __repr__(self):
        return "<{cls} {id:#016x} {name!r} {locked}>".format(
            cls=self.__class__.__name__,
            id=id(self),
            name=self._name,
            locked="locked" if self._locked else "unlocked",
        )

    def __str__(self):
        return self._name.encode("unicode_escape")

    def __unicode__(self):
        return self._name

    def __eq__(self, other):
        """Equality tests.

        >>> _View("name1") == _View("name1")
        False
        >>> t1 = t2 = _View("name3")
        >>> t1 is t2
        True
        >>> t1 == t2
        True

        Views can be compared to other views (uses ``is``),
        or by (unicode) string equality.

        """
        return self is other

    def __hash__(self):
        return id(self)

    def __bool__(self):
        return True

    def to_jsf(self):
        """Convert to a form serializable by the json module."""
        return {
            "name": self.name,
            "locked": self.locked,
        }

    @classmethod
    def new_from_jsf(cls, jsf):
        """Returns a new _View built from the json-serializable form."""
        return cls(**jsf)


class _NamedViewsSet (object):
    """A set of _View objects that enforces unique naming."""

    def __init__(self):
        super(_NamedViewsSet, self).__init__()
        self.objs = set()  # {_View}
        self.names = dict()  # {str: _View}

    def add(self, view):
        if view in self.objs:
            return
        name = lib.naming.make_unique_name(
            view.name, self.names,
            always_number=NEW_VIEW_IDENT,
        )
        assert name not in self.names
        if view.name != name:
            view.name = name
        self.names[name] = view
        self.objs.add(view)

    def remove(self, view):
        if view not in self.objs:
            return
        self.remove_by_name(view.name)
        assert view not in self.objs

    def remove_by_name(self, name):
        if name not in self.names:
            return
        view = self.names.pop(name)
        self.objs.remove(view)

    def clear(self):
        self.objs.clear()
        self.names.clear()


class LayerViewManager (object):
    """Controls which layers are visible in a document with named views."""

    _SETTINGS_KEY = "layervis"
    _SETTINGS_VIEWS_SUBKEY = "views"
    _SETTINGS_ACTIVE_VIEW_SUBKEY = "active_view"
    _SETTINGS_LAYER_VIEWS_SUBKEY = "layer_views"

    # Construction:

    def __init__(self, docmodel):
        super(LayerViewManager, self).__init__()
        self._docmodel = docmodel
        root = docmodel.layer_stack
        self._stack = docmodel.layer_stack

        # The currently active view.
        # A value of None means that no named view is active.
        self._current_view = None

        # The set of views managed by the user.
        # There may be more view objects in existence,
        # for example in the command stack.
        self._views = _NamedViewsSet()

        # Layer stack monitoring.
        root.layer_properties_changed += self._stack_layer_props_changed_cb
        root.layer_inserted += self._stack_layer_inserted_cb

        # Observe myself to change layer views when it's safe.
        self.current_view_changed += self._current_view_changed_cb

        # Save and load.
        docmodel.settings.sync_pending_changes \
            += self._doc_settings_sync_pending_changes_cb
        docmodel.settings.modified += self._doc_settings_modified_cb

    # Loading and saving via the doc settings dictionary:

    def _doc_settings_sync_pending_changes_cb(self, settings, flush=False):
        """Save to the doc settings when needed (e.g. before ORA save)

        This is called before the document is saved (and at many other
        times, including autosave), so use it to store a serializable
        copy of the running state into the settings dict.

        See also: _doc_settings_modified_cb().

        """
        # Don't do anything while the doc settings are being updated.
        if self._docmodel.settings.modified.calling_observers:
            return
        # Only serious changes.
        if not flush:
            return
        # TODO: use a dirty flag to avoid too many walks?

        # The user-managed list of views
        views_list = [v.to_jsf() for v in self._views.objs]

        # The currently active view
        current = None
        if self._current_view is not None:
            current = self._current_view.name

        # Each layer's views
        by_path = []
        for path, layer in self._docmodel.layer_stack.walk():
            vset = self._get_vset_for_layer(layer)
            names = [v.name for v in vset if v in self._views.objs]
            by_path.append([list(path), names])

        self._docmodel.settings[self._SETTINGS_KEY] = {
            self._SETTINGS_VIEWS_SUBKEY: views_list,
            self._SETTINGS_ACTIVE_VIEW_SUBKEY: current,
            self._SETTINGS_LAYER_VIEWS_SUBKEY: by_path,
        }

    def _doc_settings_modified_cb(self, settings, oldvalues):
        """Update state when the doc settings change (e.g. ORA load).

        This clears and completely rebuilds the internal state of the
        manager object to match the serializable form in the settings
        dict. An important use is to restore the state after the
        document settings have been loaded from a .ora file.

        See also: _doc_sync_pending_changes_cb().

        """
        # Don't run when storing.
        if self._docmodel.settings.sync_pending_changes.calling_observers:
            return
        # Only when the settings key changes.
        if self._SETTINGS_KEY not in oldvalues:
            return

        sdict = settings.get(self._SETTINGS_KEY, {})

        # Reconstruct the user-managed list of views from a list of
        # stored names.
        self._views.clear()
        try:
            views_list = sdict.get(self._SETTINGS_VIEWS_SUBKEY, [])
            for jsf in views_list:
                view = _View.new_from_jsf(jsf)
                self._views.add(view)
        except Exception:
            logger.exception(
                "settings-modified: failed to load settings[%r][%r]",
                self._SETTINGS_KEY,
                self._SETTINGS_VIEWS_SUBKEY,
            )

        # Reconstruct the current view from the stored name.
        self._current_view = None
        try:
            current_name = sdict.get(self._SETTINGS_ACTIVE_VIEW_SUBKEY, None)
            if current_name is not None:
                self._current_view = self._views.names.get(current_name, None)
        except Exception:
            logger.exception(
                "settings-modified: failed to load settings[%r][%r]",
                self._SETTINGS_KEY,
                self._SETTINGS_ACTIVE_VIEW_SUBKEY,
            )

        # Reconstruct layer vsets from he stored names.
        path_to_layer = {}
        for path, layer in self._docmodel.layer_stack.walk():
            path_to_layer[path] = layer
            vset = self._get_vset_for_layer(layer)
            vset.clear()
        try:
            vset_rec = sdict.get(self._SETTINGS_LAYER_VIEWS_SUBKEY, [])
            for path, tagnames in vset_rec:
                path = tuple(path)
                layer = path_to_layer.get(path)
                if layer is None:
                    return
                layer_vset = self._get_vset_for_layer(layer)
                for tagname in tagnames:
                    view = self._views.names.get(tagname, None)
                    if tagname is None:
                        continue
                    layer_vset.add(view)
        except Exception:
            logger.exception(
                "settings-modified: failed to load [%r]['layer-visibs']",
                self._SETTINGS_KEY,
            )

        # Announce the changes, to update any connected UI.
        self.view_names_changed()
        self.current_view_changed()

    # Public properties

    @property
    def current_view_name(self):
        """RO property: the current view's name.

        :returns: The name of the current view, or None
        :rtype: unicode

        If the current view name is None, the current view is the
        built-in unnamed and unsaved view.

        """
        if self._current_view is None:
            return None
        return self._current_view.name

    @property
    def current_view_locked(self):
        """RO property: the current view's lock state.

        :returns: Whether the current view is locked.
        :rtype: bool

        The built-in, unsaved view is always unlocked.

        """
        if self._current_view is None:
            return False
        return self._current_view.locked

    @property
    def view_names(self):
        """RO property: list of the current set of managed view names."""
        return list(self._views.names.keys())

    # Public events:

    @event
    def current_view_changed(self):
        """Event: the current view was changed."""

    @event
    def view_names_changed(self):
        """Event: view_names is different."""

    # Self-observation:

    def _current_view_changed_cb(self, _lvm):
        """Respond to a change of the current view: set layer visibilities.

        Doing this as a self-observer callback method allows
        _stack_layer_props_changed_cb to safely ignore the changes to
        layer visibilities that originate here.

        """
        view = self._current_view
        if view is None:
            return

        for path, layer in self._stack.walk():
            vset = self._get_vset_for_layer(layer)
            if (view in vset) and not layer.visible:
                layer.visible = True
            elif (view not in vset) and layer.visible:
                layer.visible = False

    # Document observation:

    def _stack_layer_props_changed_cb(self, root, path, layer, changed):
        """Respond to any outside change of layer "visible" properties."""
        view = self._current_view
        if view is None:
            return
        if self.current_view_changed.calling_observers:
            return
        if "visible" not in changed:
            return
        vset = self._get_vset_for_layer(layer)
        if layer.visible:
            vset.add(view)
        elif view in vset:
            vset.remove(view)

    def _stack_layer_inserted_cb(self, root, path):
        view = self._current_view
        if view is None:
            return
        layer = root.deepget(path)
        if layer is None:
            return
        vset = self._get_vset_for_layer(layer)
        if layer.visible:
            vset.add(view)
        elif view in vset:
            vset.remove(view)

    def _get_vset_for_layer(self, layer):
        """Gets the "visible in views" set for a layer."""
        try:
            vset = layer.__visible_in_views
        except AttributeError:
            vset = set()
            layer.__visible_in_views = vset
        return vset

    # Higher-level API:

    def clear(self):
        self._current_view = None
        self.current_view_changed()
        self._views.clear()
        self.view_names_changed()

    def add_new_view(self, name=None):
        """Adds a new named view capturing the currently visible layers.

        :param unicode name: Base name for a new named view, or None.
        :rtype: _View
        :returns: the added view.

        If name=None or name="" is passed, the new view will be named
        uniquely after NEW_VIEW_IDENT. The None value is reserved for
        representing the default working view.

        """
        if name is None or name == "":
            name = NEW_VIEW_IDENT
        name = unicode(name)

        # All currently visible layers are tagged as visible in the new view.
        view = _View(name)
        for path, layer in self._stack.walk():
            if layer.visible:
                vset = self._get_vset_for_layer(layer)
                vset.add(view)

        self._views.add(view)
        self.view_names_changed()
        self.activate_view(view)
        return view

    def add_view(self, view):
        """Adds a view, but does not activate it."""
        if view is None:
            raise ValueError("Cannot add None")
        self._views.add(view)
        self.view_names_changed()

    def activate_view(self, view):
        """Activates a view."""
        if view is not None:
            if view not in self._views.objs:
                raise ValueError("view not in views list")
        self._current_view = view
        self.current_view_changed()

    def activate_view_by_name(self, name):
        """Activates a view by name."""
        view = None
        if name is not None:
            view = self._views.names.get(name, None)
            if view is None:
                raise ValueError("no view named %r in views list" % (name,))
        return self.activate_view(view)

    def remove_active_view(self, restore=None):
        """Removes the currently active view.

        :param _View restore: Optional view to restore.

        """
        view = self._current_view
        if view is None:
            raise ValueError("Cannot delete the default view.")

        self.activate_view(restore)

        self._views.remove(view)
        self.view_names_changed()

        return view

    def rename_active_view(self, name):
        """Renames the currently active view.

        :param unicode name: Base name for the new named view.
        :rtype: tuple
        :returns: The old and new names, as (old_name, new_unique_name)

        """
        view = self._current_view
        if view is None:
            raise ValueError("Cannot rename the default view.")
        if name == "" or name is None:
            name = NEW_VIEW_IDENT
        old_name = view.name
        popped_tag = self._views.names.pop(old_name)
        assert popped_tag is view
        new_name = lib.naming.make_unique_name(
            name, self._views.names,
            always_number=NEW_VIEW_IDENT,
        )
        self._views.names[new_name] = popped_tag
        view.name = new_name

        self.view_names_changed()
        self.current_view_changed()

        return (old_name, new_name)

    def set_active_view_locked(self, locked):
        view = self._current_view
        if view is None:
            raise ValueError("Cannot lock or unlock the default view.")
        locked = bool(locked)
        if locked != bool(view.locked):
            view.locked = locked
            self.current_view_changed()


# Command classes:

class AddLayerView (Command):
    """Adds a new layer visibility set, capturing what's visible now."""

    def __init__(self, doc, name=None, **kwds):
        super(AddLayerView, self).__init__(doc, name=name, **kwds)
        self._lvm = doc.layer_view_manager
        self._prev_active_view = None
        self._new_view_name_orig = name

    def redo(self):
        assert self._prev_active_view is None
        self._prev_active_view = self._lvm._current_view
        name = self._new_view_name_orig
        self._lvm.add_new_view(name=name)
        assert self._lvm._current_view is not self._prev_active_view
        assert self._lvm._current_view is not None

    def undo(self):
        assert self._lvm._current_view is not None
        self._lvm.remove_active_view(restore=self._prev_active_view)
        assert self._lvm._current_view is self._prev_active_view
        self._prev_active_view = None

    @property
    def display_name(self):
        return C_(
            "layer views: commands: add",
            u"Add Layer View",
        )


class RemoveActiveLayerView (Command):
    """Removes the active layer view."""

    def __init__(self, doc, **kwds):
        super(RemoveActiveLayerView, self).__init__(doc, **kwds)
        self._lvm = doc.layer_view_manager
        self._removed_view = None

    def redo(self):
        removed = self._lvm.remove_active_view()
        self._removed_view = removed

    def undo(self):
        self._lvm.add_view(self._removed_view)
        self._lvm.activate_view(self._removed_view)
        self._removed_view = None

    @property
    def display_name(self):
        return C_(
            "layer views: commands: remove",
            u"Remove Layer View",
        )


class RenameActiveLayerView (Command):
    """Renames the active layer view."""

    def __init__(self, doc, name, **kwds):
        super(RenameActiveLayerView, self).__init__(doc, name=name, **kwds)
        self._lvm = doc.layer_view_manager

        self._old_name = None
        self._new_name = name

    def redo(self):
        assert self._old_name is None
        (old_name, new_name) = self._lvm.rename_active_view(self._new_name)
        self._old_name = old_name

    def undo(self):
        assert self._old_name is not None
        old_name = self._old_name
        self._lvm.rename_active_view(old_name)
        self._old_name = None

    @property
    def display_name(self):
        return C_(
            "layer views: commands: remove",
            u"Remove Layer View",
        )


class ActivateLayerView (Command):
    """Makes a different layer view active."""

    def __init__(self, doc, name, **kwds):
        super(ActivateLayerView, self).__init__(doc, name=name, **kwds)
        self._lvm = doc.layer_view_manager
        assert name in self._lvm.view_names or name is None
        assert name is not self._lvm.current_view_name
        self._prev_view = None
        self._next_view_name = name

    def redo(self):
        prev = self._lvm._current_view
        self._lvm.activate_view_by_name(self._next_view_name)
        self._prev_view = prev

    def undo(self):
        prev = self._prev_view
        self._lvm.activate_view(prev)
        assert prev is self._lvm._current_view
        self._prev_view = None

    @property
    def display_name(self):
        return C_(
            "layer views: commands: activate",
            u"Activate Layer View “{name}”",
        ).format(
            name=self._next_view_name,
        )


class SetActiveLayerViewLocked (Command):
    """Sets the locked state of the active layer view."""

    def __init__(self, doc, locked, **kwds):
        super(SetActiveLayerViewLocked, self) \
            .__init__(doc, locked=locked, **kwds)
        self._lvm = doc.layer_view_manager
        self._old_locked = None
        self._new_locked = locked

    def redo(self):
        self._old_locked = self._lvm.current_view_locked
        self._lvm.set_active_view_locked(self._new_locked)

    def undo(self):
        self._lvm.set_active_view_locked(self._old_locked)
        self._old_locked = None

    @property
    def display_name(self):
        if self._new_locked:
            return C_(
                "layer views: commands: set active view's lock flag",
                u"Lock Layer View",
            )
        else:
            return C_(
                "layer views: commands: set active view's lock flag",
                u"Unlock Layer View",
            )



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Brush color changer."""

from __future__ import division, print_function

from . import colors
import lib.color


class BrushColorManager (colors.ColorManager):
    """Color manager mediating between brush settings and the color adjusters.
    """

    __brush = None
    __in_callback = False

    def __init__(self, app):
        """Initialize, binding to certain events.
        """
        colors.ColorManager.__init__(self, app.preferences, app.datapath)
        self.__brush = app.brush
        app.brush.observers.append(self.__settings_changed_cb)
        app.doc.input_stroke_ended += self.__input_stroke_ended_cb
        app.doc.model.sync_pending_changes += self.__sync_pending_changes_cb
        self._app = app

    def set_color(self, color):
        """Propagate user-set colors to the brush too (extension).
        """
        colors.ColorManager.set_color(self, color)
        if not self.__in_callback:
            self.__brush.set_color_hsv(color.get_hsv())

    def __settings_changed_cb(self, settings):
        # When the color changes by external means, update the adjusters.
        if not settings.intersection(('color_h', 'color_s', 'color_v')):
            return
        brush_color = lib.color.HSVColor(*self.__brush.get_color_hsv())
        if brush_color == self.get_color():
            return
        self.__in_callback = True
        self.set_color(brush_color)
        self.__in_callback = False

    def __input_stroke_ended_cb(self, doc, event):
        # Update the color usage history immediately after the user
        # makes an explicit (pen down/stroke/up) brushstroke,
        # for responsiveness.
        brush = self.__brush
        if not brush.is_eraser():
            col = lib.color.HSVColor(*brush.get_color_hsv())
            self.push_history(col)

    def __sync_pending_changes_cb(self, model, flush=True, **kwargs):
        # Update the color usage history after any flushing sync.
        #
        # Freehand mode sends a flushing sync request after painting
        # continuously for some seconds if pixel changes were made.
        # Pixel changes can happen even without pressure so we need
        # this hook as well as __input_stroke_ended_cb for correctness.
        if not flush:
            return
        brush = self.__brush
        if not brush.is_eraser():
            col = lib.color.HSVColor(*brush.get_color_hsv())
            self.push_history(col)



# This file is part of MyPaint.
# Copyright (C) 2017 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Interfaces and objects for providing UI feedback."""


# Imports:

from __future__ import division, print_function

from lib.observable import event
import lib.helpers


# Class defs:

class Progress (object):
    """Itemized progress report with hierarchy.

    A top-level GUI process can create one of these objects, and connect
    to its .changed event to get real-time reports from calls it passes
    the Progress down to.

    >>> prog0 = Progress()
    >>> feedback = []
    >>> prog0.changed += lambda p: feedback.append(p.fraction)

    If the number of items is unset, the fraction is None.  Code that
    doesn't know how much work it has to do can call changed() to notify
    observers, however. This may cause a pulse in GUI progress bars, for
    example.

    >>> prog0.fraction is None
    True
    >>> prog0.changed()
    >>> prog0.changed()
    >>> prog0.fraction is None
    True
    >>> feedback
    [None, None]

    Progress report objects are only really useful if each stage
    declares how many items it expects to be filling in. You can't set
    the item count to less than its current value, and each change to ie
    emits the changed() event.

    >>> prog0.items = 10
    >>> prog0.fraction
    0.0
    >>> feedback
    [None, None, 0.0]

    You can increment the progress report to mark items complete,
    or explicitly declare how many items are now complete.

    >>> prog0 += 1
    >>> prog0 += 2
    >>> feedback
    [None, None, 0.0, 0.1, 0.3]
    >>> prog0.completed(5)
    >>> feedback
    [None, None, 0.0, 0.1, 0.3, 0.5]

    If a stage is more fine-grained, you can open() a sub-progress
    report and pass that around or fill it in.

    >>> p1 = prog0.open()
    >>> p2 = prog0.open()
    >>> p1.items = 2
    >>> p2.items = 4
    >>> feedback
    [None, None, 0.0, 0.1, 0.3, 0.5, 0.5, 0.5]
    >>> prog0.fraction
    0.5
    >>> p1 += 1
    >>> p2 += 1
    >>> prog0.fraction
    0.575
    >>> feedback
    [None, None, 0.0, 0.1, 0.3, 0.5, 0.5, 0.5, 0.55, 0.575]

    """

    # Initialization and basic fields:

    def __init__(self):
        """Create a new, unsized Progress freedback object.

        It is reasonable for functions supporting this API to demand
        either unsized or sized Progress objects.

        """
        super(Progress, self).__init__()
        self._items = None
        self._completed = 0
        self._open = {}  # {Progress: weight_in_items}

    @property
    def items(self):
        """The number of open items (read, write once).

        An attempt to set the number of items after it has already been
        set results in an exception.

        >>> p = Progress()
        >>> p.items = 10
        >>> p.items = 15   # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: ...

        """
        return self._items

    @items.setter
    def items(self, n):
        if self._items is not None:
            raise ValueError("Progress.items has already been set.")
        n = max(0, int(n))
        if (self._items is None) or (n > self._items):
            self._items = n
            self.changed()

    # Observable events and monitoring:

    @event
    def changed(self):
        """Event: fires when "fraction" changes."""

    def _child_changed_cb(self, sprog):
        """Handles changes to an open child Progress."""
        if sprog not in self._open:
            return
        if self._items is None:
            self.changed()
            return
        if sprog.fraction >= 1:
            sprog_weight = self._open.pop(sprog)
            self._completed += sprog_weight
            sprog.changed -= self._child_changed_cb
        self.changed()

    # Casts:

    def __bool__(self):
        """Progress objects always test true.

        >>> p = Progress()
        >>> bool(p)
        True

        """
        return True

    def __int__(self):
        """The integer value is the number of completed items.

        >>> p = Progress()
        >>> p.items = 10
        >>> p += 3
        >>> p1 = p.open()
        >>> p1.items = 97
        >>> p1.close()
        >>> int(p)
        4

        """
        return self._completed

    # Updating of open WIP objs:

    def __iadd__(self, n):
        """Mark this many items complete.

        Internally uses completed(), so it is subject to the same
        constraints.

        """
        n = max(0, int(n))
        c = self._completed + n
        self.completed(c)
        return self

    def completed(self, c):
        """Try to mark up to item c as completed.

        You cannot mark as complete more than ``items`` minus the
        combined weight of all the currently open child Progress
        objects.

        """
        if self._items is None:
            self.changed()
            return
        c = lib.helpers.clamp(c, 0, (self._items - self._open_items_weight))
        if c > self._completed:
            self._completed = c
            self.changed()

    def close(self):
        """Mark this Progress as complete.

        If you set the size of a Progress object in a given function,
        and start marking its items as complete, it's normal to close()
        the Progress object later on.

        Or to put it another way, parent function calls male new,
        unsized Progress objects with open(), and pass those to
        functions they invoke. The child functions set sizes, declare
        that work is done in an incremental fashion, and finally call
        close() on what they're passed.

        """
        if self._items is not None:
            self._completed = self._items
        for sprog in self._open.keys():
            sprog.changed -= self._child_changed_cb
        self._open.clear()
        self.changed()

    def __repr__(self):
        """String representation of a Progress object.

        >>> p = Progress()
        >>> p
        <Progress 0.0/None>
        >>> p.items = 10
        >>> p += 1
        >>> p
        <Progress 1.0/10>

        """
        repr_str = "<Progress %0.1f/%r>" % (
            self._completed + self._open_items_completion,
            self._items,
        )
        return repr_str

    @property
    def _open_items_weight(self):
        """Total weight of all open items (monitored child objs)."""
        total = 0
        for p, w in self._open.items():
            total += w
        return total

    @property
    def _open_items_completion(self):
        """Weighted completion sum over all open items."""
        total = 0.0
        for p, w in self._open.items():
            f = p.fraction
            if f is None:
                f = 0.0
            total += f * float(w)
        return total

    @property
    def fraction(self):
        """Read-only completeness fraction."""
        if self._items is None:
            return None
        if self._items <= 0:
            return 1.0
        f1 = float(self._completed)
        f1 += self._open_items_completion
        f1 /= float(self._items)
        return lib.helpers.clamp(f1, 0.0, 1.0)

    def open(self, weight=1):
        """Open a new monitored child Progress.

        :param int weight: Number of items represented by the child.
        :returns: A new, unsized Progress object.
        :rtype: Progress

        A child Progress object cannot represent more items than are
        currently available. If you try, the returned Progress object
        will work, but it won't be monitored.

        >>> p = Progress()
        >>> p.items = 10
        >>> p1 = p.open(4)
        >>> p1.items = 2
        >>> p1 += 1
        >>> p
        <Progress 2.0/10>
        >>> p1.close()
        >>> p
        <Progress 4.0/10>

        """
        sprog_weight = max(1, int(weight))
        sprog = Progress()
        if self._items is not None:
            c = self._completed + self._open_items_weight + sprog_weight
            if c < self._items:
                sprog.changed += self._child_changed_cb
                self._open[sprog] = sprog_weight
        return sprog


def _test():
    """Run doctests"""
    import doctest
    doctest.testmod()


if __name__ == '__main__':
    _test()



# This file is part of MyPaint.
# Copyright (C) 2014-2017 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Layer manipulation GUI helper code"""


## Imports

from __future__ import division, print_function

import lib.layer
from lib.xml import escape
from lib.observable import event
from lib import helpers

from lib.document import Document
from lib.gettext import gettext as _
from lib.gettext import C_
from gui.layerprops import make_preview
import gui.drawutils
from lib.pycompat import unicode

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GObject
from lib.gibindings import GLib
from lib.gibindings import Pango
from lib.gibindings import GdkPixbuf

import sys
import logging


## Module vars

logger = logging.getLogger(__name__)


## Class defs


class RootStackTreeModelWrapper (GObject.GObject, Gtk.TreeModel):
    """Tree model wrapper presenting a document model's layers stack

    Together with the layers panel (defined in `gui.layerswindow`),
    and `RootStackTreeView`, this forms part of the presentation logic
    for the layer stack.

    """

    ## Class vars

    INVALID_STAMP = 0
    MIN_VALID_STAMP = 1
    COLUMN_TYPES = (object,)
    LAYER_COLUMN = 0

    ## Setup

    def __init__(self, docmodel):
        """Initialize, presenting the root stack of a document model

        :param Document docmodel: model to present
        """
        super(RootStackTreeModelWrapper, self).__init__()
        self._docmodel = docmodel
        root = docmodel.layer_stack
        self._root = root
        self._iter_stamp = 1
        self._iter_id2path = {}  # {pathid: pathtuple}
        self._iter_path2id = {}   # {pathtuple: pathid}
        root.layer_properties_changed += self._layer_props_changed_cb
        root.layer_inserted += self._layer_inserted_cb
        root.layer_deleted += self._layer_deleted_cb
        root.layer_thumbnail_updated += self._layer_thumbnail_updated_cb
        lvm = docmodel.layer_view_manager
        lvm.current_view_changed += self._lvm_current_view_changed_cb
        self._drag = None

    ## Python boilerplate

    def __repr__(self):
        nrows = len(list(self._root.deepiter()))
        return "<%s n=%d>" % (self.__class__.__name__, nrows)

    ## Event and update handling

    def _layer_props_changed_cb(self, root, layerpath, layer, changed):
        """Updates the display after a layer's properties change"""
        treepath = Gtk.TreePath(layerpath)
        it = self.get_iter(treepath)
        self._row_changed_all_descendents(treepath, it)

    def _layer_thumbnail_updated_cb(self, root, layerpath, layer):
        """Updates the display after a layer's thumbnail changes."""
        treepath = Gtk.TreePath(layerpath)
        it = self.get_iter(treepath)
        self.row_changed(treepath, it)

    def _layer_inserted_cb(self, root, path):
        """Updates the display after a layer is added"""
        self.invalidate_iters()
        it = self.get_iter(path)
        self.row_inserted(Gtk.TreePath(path), it)
        parent_path = path[:-1]
        if not parent_path:
            return
        parent = self._root.deepget(parent_path)
        if len(parent) == 1:
            parent_it = self.get_iter(parent_path)
            self.row_has_child_toggled(Gtk.TreePath(parent_path),
                                       parent_it)

    def _layer_deleted_cb(self, root, path):
        """Updates the display after a layer is removed"""
        self.invalidate_iters()
        self.row_deleted(Gtk.TreePath(path))
        parent_path = path[:-1]
        if not parent_path:
            return
        parent = self._root.deepget(parent_path)
        if len(parent) == 0:
            parent_it = self.get_iter(parent_path)
            self.row_has_child_toggled(Gtk.TreePath(parent_path),
                                       parent_it)

    def _row_dragged(self, src_path, dst_path):
        """Handles the user dragging a row to a new location"""
        self._docmodel.restack_layer(src_path, dst_path)

    def _row_changed_all_descendents(self, treepath, it):
        """Like GtkTreeModel.row_changed(), but all descendents too."""
        self.row_changed(treepath, it)
        if self.iter_n_children(it) <= 0:
            return
        ci = self.iter_nth_child(it, 0)
        while ci is not None:
            treepath = self.get_path(ci)
            self._row_changed_all_descendents(treepath, ci)
            ci = self.iter_next(ci)

    def _row_changed_all(self):
        """Like GtkTreeModel.row_changed(), but all rows."""
        it = self.get_iter_first()
        while it is not None:
            treepath = self.get_path(it)
            self._row_changed_all_descendents(treepath, it)
            it = self.iter_next(it)

    def _lvm_current_view_changed_cb(self, lvm):
        """Respond to changes of/on the currently active layer-view.

        For the sake of the related TreeView, announce a change to all
        rows to make sure any bulk changes to the sensitive state of the
        visibility column are visible instantly.

        This is slightly incorrect, since it means that the TreeModel
        needs to know what its TreeView does. Maybe the model
        implemented here should expose its data in proper columns, with
        effective-visibility, visibility-sensitive and so on.

        """
        self._row_changed_all()

    ## Iterator management

    def invalidate_iters(self):
        """Invalidates all iters produced by this model"""
        # No need to zap the lookup tables: tree paths have a tightly
        # controlled vocabulary.
        if self._iter_stamp == sys.maxsize:
            self._iter_stamp = self.MIN_VALID_STAMP
        else:
            self._iter_stamp += 1

    def iter_is_valid(self, it):
        """True if an iterator produced by this model is valid"""
        return it.stamp == self._iter_stamp

    @classmethod
    def _invalidate_iter(cls, it):
        """Invalidates an iterator"""
        it.stamp = cls.INVALID_STAMP
        it.user_data = None

    def _get_iter_path(self, it):
        """Gets an iterator's path: None if invalid"""
        if not self.iter_is_valid(it):
            return None
        else:
            path = self._iter_id2path.get(it.user_data)
            return tuple(path)

    def _set_iter_path(self, it, path):
        """Sets an iterator's path, invalidating it if path=None"""
        if path is None:
            self._invalidate_iter(it)
        else:
            it.stamp = self._iter_stamp
            pathid = self._iter_path2id.get(path)
            if pathid is None:
                path = tuple(path)
                pathid = id(path)
                self._iter_path2id[path] = pathid
                self._iter_id2path[pathid] = path
            it.user_data = pathid

    def _create_iter(self, path):
        """Creates an iterator for the given path

        The returned pair can be returned by the ``do_*()`` virtual
        function implementations. Use this method in preference to the
        regular `Gtk.TreeIter` constructor.
        """
        if not path:
            return (False, None)
        else:
            it = Gtk.TreeIter()
            self._set_iter_path(it, tuple(path))
            return (True, it)

    def _iter_bump(self, it, delta):
        """Move an iter at its current level"""
        path = self._get_iter_path(it)
        if path is not None:
            path = list(path)
            path[-1] += delta
            path = tuple(path)
        if self.get_layer(treepath=path) is None:
            self._invalidate_iter(it)
            return False
        else:
            self._set_iter_path(it, path)
            return True

    ## Data lookup

    def get_layer(self, treepath=None, it=None):
        """Look up a layer using paths or iterators"""
        if treepath is None:
            if it is not None:
                treepath = self._get_iter_path(it)
        if treepath is None:
            return None
        if isinstance(treepath, Gtk.TreePath):
            treepath = tuple(treepath.get_indices())
        return self._root.deepget(treepath)

    ## GtkTreeModel vfunc implementation

    def do_get_flags(self):
        """Fetches GtkTreeModel flags"""
        return 0

    def do_get_n_columns(self):
        """Count of GtkTreeModel columns"""
        return len(self.COLUMN_TYPES)

    def do_get_column_type(self, n):
        return self.COLUMN_TYPES[n]

    def do_get_iter(self, treepath):
        """New iterator pointing at a node identified by GtkTreePath"""
        if not self.get_layer(treepath=treepath):
            treepath = None
        return self._create_iter(treepath)

    def do_get_path(self, it):
        """New GtkTreePath for a treeiter"""
        path = self._get_iter_path(it)
        if path is None:
            return None
        else:
            return Gtk.TreePath(path)

    def do_get_value(self, it, column):
        """Value at a particular row-iterator and column index"""
        if column != 0:
            return None
        return self.get_layer(it=it)

    def do_iter_next(self, it):
        """Move an iterator to the node after it, returning success"""
        return self._iter_bump(it, 1)

    def do_iter_previous(self, it):
        """Move an iterator to the node before it, returning success"""
        return self._iter_bump(it, -1)

    def do_iter_children(self, parent):
        """Fetch an iterator pointing at the first child of a parent"""
        return self.do_iter_nth_child(parent, 0)

    def do_iter_has_child(self, it):
        """True if an iterator has children"""
        layer = self.get_layer(it=it)
        return isinstance(layer, lib.layer.LayerStack) and len(layer) > 0

    def do_iter_n_children(self, it):
        """Count of the children of a given iterator"""
        layer = self.get_layer(it=it)
        if not isinstance(layer, lib.layer.LayerStack):
            return 0
        else:
            return len(layer)

    def do_iter_nth_child(self, it, n):
        """Fetch a specific child iterator of a parent iter"""
        if it is None:
            path = (n,)
        else:
            path = self._get_iter_path(it)
            if path is not None:
                path = list(self._get_iter_path(it))
                path.append(n)
                path = tuple(path)
        if not self.get_layer(treepath=path):
            path = None
        return self._create_iter(path)

    def do_iter_parent(self, it):
        """Fetches the parent of a valid iterator"""
        if it is None:
            parent_path = None
        else:
            path = self._get_iter_path(it)
            if path is None:
                parent_path = None
            else:
                parent_path = list(path)
                parent_path.pop(-1)
                parent_path = tuple(parent_path)
            if parent_path == ():
                parent_path = None
        return self._create_iter(parent_path)


class RootStackTreeView (Gtk.TreeView):
    """GtkTreeView tailored for a doc's root layer stack"""

    DRAG_HOVER_EXPAND_TIME = 1.25   # seconds

    def __init__(self, docmodel):
        super(RootStackTreeView, self).__init__()
        self._docmodel = docmodel

        treemodel = RootStackTreeModelWrapper(docmodel)
        self.set_model(treemodel)

        target1 = Gtk.TargetEntry.new(
            target = "GTK_TREE_MODEL_ROW",
            flags = Gtk.TargetFlags.SAME_WIDGET,
            info = 1,
        )
        self.drag_source_set(
            start_button_mask = Gdk.ModifierType.BUTTON1_MASK,
            targets = [target1],
            actions = Gdk.DragAction.MOVE,
        )
        self.drag_dest_set(
            flags = Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP,
            targets = [target1],
            actions = Gdk.DragAction.MOVE,
        )

        self.connect("button-press-event", self._button_press_cb)

        # Motion and modifier keys during drag
        self.connect("drag-begin", self._drag_begin_cb)
        self.connect("drag-motion", self._drag_motion_cb)
        self.connect("drag-leave", self._drag_leave_cb)
        self.connect("drag-drop", self._drag_drop_cb)
        self.connect("drag-end", self._drag_end_cb)

        # Track updates from the model
        self._processing_model_updates = False
        root = docmodel.layer_stack
        root.current_path_updated += self._current_path_updated_cb
        root.expand_layer += self._expand_layer_cb
        root.collapse_layer += self._collapse_layer_cb
        root.layer_content_changed += self._layer_content_changed_cb
        root.current_layer_solo_changed += lambda *a: self.queue_draw()

        # View behaviour and appearance
        self.set_headers_visible(False)
        selection = self.get_selection()
        selection.set_mode(Gtk.SelectionMode.BROWSE)
        self.set_size_request(150, 200)

        # Visibility flag column
        col = Gtk.TreeViewColumn(_("Visible"))
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self._flags1_col = col

        # Visibility cell
        cell = Gtk.CellRendererPixbuf()
        col.pack_start(cell, False)
        datafunc = self._layer_visible_pixbuf_datafunc
        col.set_cell_data_func(cell, datafunc)

        # Name and preview column: will be indented
        col = Gtk.TreeViewColumn(_("Name"))
        col.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)
        self._name_col = col

        # Preview cell
        cell = Gtk.CellRendererPixbuf()
        col.pack_start(cell, False)
        datafunc = self._layer_preview_pixbuf_datafunc
        col.set_cell_data_func(cell, datafunc)
        self._preview_cell = cell

        # Name cell
        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        col.pack_start(cell, True)
        datafunc = self._layer_name_text_datafunc
        col.set_cell_data_func(cell, datafunc)
        col.set_expand(True)
        col.set_min_width(48)

        # Other flags column
        col = Gtk.TreeViewColumn(_("Flags"))
        col.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)
        area = col.get_property("cell-area")
        area.set_orientation(Gtk.Orientation.VERTICAL)
        self._flags2_col = col

        # Locked cell
        cell = Gtk.CellRendererPixbuf()
        col.pack_end(cell, False)
        datafunc = self._layer_locked_pixbuf_datafunc
        col.set_cell_data_func(cell, datafunc)

        # Column order on screen
        self._columns = [
            self._flags1_col,
            self._name_col,
            self._flags2_col,
        ]
        for col in self._columns:
            self.append_column(col)

        # View appearance
        self.set_show_expanders(True)
        self.set_enable_tree_lines(True)
        self.set_expander_column(self._name_col)

        self.connect_after("show", self._post_show_cb)

    ## Low-level GDK event handlers

    def _button_press_cb(self, view, event):
        """Handle button presses (visibility, locked, naming)"""

        # Basic details about the click
        single_click = (event.type == Gdk.EventType.BUTTON_PRESS)
        double_click = (event.type == Gdk.EventType._2BUTTON_PRESS)
        is_menu = event.triggers_context_menu()

        # Determine which row & column was clicked
        x, y = int(event.x), int(event.y)
        bw_x, bw_y = view.convert_widget_to_bin_window_coords(x, y)
        click_info = view.get_path_at_pos(bw_x, bw_y)
        if click_info is None:
            return True
        treemodel = self.get_model()
        click_treepath, click_col, cell_x, cell_y = click_info
        click_layer = treemodel.get_layer(treepath=click_treepath)
        click_layerpath = tuple(click_treepath.get_indices())

        # Defer certain kinds of click to separate handlers. These
        # handlers can return True to stop processing and indicate that
        # the current layer should not be changed.
        col_handlers = [
            # (Column, CellRenderer, single, double, handler)
            (self._flags1_col, None, True, False,
             self._flags1_col_click_cb),
            (self._flags2_col, None, True, False,
             self._flags2_col_click_cb),
            (self._name_col, None, False, True,
             self._name_col_2click_cb),
            (self._name_col, self._preview_cell, True, False,
             self._preview_cell_click_cb),
        ]
        if not is_menu:
            for col, cell, when_single, when_double, handler in col_handlers:
                if when_single and not single_click:
                    continue
                if when_double and not double_click:
                    continue
                # Correct column?
                if col is not click_col:
                    continue
                # Click inside the target column's entire area?
                ca = view.get_cell_area(click_treepath, col)
                if not (ca.x <= bw_x < (ca.x + ca.width)):
                    continue
                # Also, inside any target CellRenderer's area?
                if cell:
                    pos_info = col.cell_get_position(cell)
                    cell_xoffs, cell_w = pos_info
                    if None in (cell_xoffs, cell_w):
                        continue
                    cell_x = ca.x + cell_xoffs
                    if not (cell_x <= bw_x < (cell_x + cell_w)):
                        continue
                # Run the delegated handler if we got here.
                if handler(event, click_layer, click_layerpath, ca):
                    return True

        # Clicks that fall thru the above cause a layer change.
        if click_layerpath != self._docmodel.layer_stack.current_path:
            self._docmodel.select_layer(path=click_layerpath)
            self.current_layer_changed()

        # Context menu for the layer just (right) clicked.
        if is_menu and single_click:
            self.current_layer_menu_requested(event)
            return True

        # Default behaviours: allow expanders & drag-and-drop to work
        return False

    def _name_col_2click_cb(self, event, layer, path, area):
        """Rename the current layer."""
        # At this point, a layer will have already been selected by
        # a single-click event.
        self.current_layer_rename_requested()
        return True

    def _flags1_col_click_cb(self, event, layer, path, area):
        """Toggle visibility or Layer Solo (with Ctrl held)."""
        rootstack = self._docmodel.layer_stack
        lvm = self._docmodel.layer_view_manager

        # Always turn off solo mode, if it's on.
        if rootstack.current_layer_solo:
            rootstack.current_layer_solo = False

        # Use Ctrl+click to torn solo mode on.
        elif event.state & Gdk.ModifierType.CONTROL_MASK:
            rootstack.current_layer_solo = True

        # Normally, clicks set the layer visible state.
        # The view can be locked elsewhere, which stops this.
        elif not lvm.current_view_locked:
            new_visible = not layer.visible
            self._docmodel.set_layer_visibility(new_visible, layer)

        return True

    def _flags2_col_click_cb(self, event, layer, path, area):
        """Toggle the clicked layer's visibility."""
        new_locked = not layer.locked
        self._docmodel.set_layer_locked(new_locked, layer)
        return True

    def _preview_cell_click_cb(self, event, layer, path, area):
        """Expand the clicked layer if the preview is clicked."""
        # The idea here is that the preview cell area acts as an extra
        # expander. Some themes' expander arrows are very small.
        treepath = Gtk.TreePath(path)
        self.expand_to_path(treepath)
        return False  # fallthru: allow the layer to be selected

    def _drag_begin_cb(self, view, context):
        self.drag_began()
        src_path = self._docmodel.layer_stack.get_current_path()
        self._drag_src_path = src_path
        self._drag_dest_path = None
        src_treepath = Gtk.TreePath(src_path)
        src_icon_surf = self.create_row_drag_icon(src_treepath)
        Gtk.drag_set_icon_surface(context, src_icon_surf)
        self._hover_expand_timer_id = None

    def _get_checked_dest_row_at_pos(self, x, y):
        """Like get_dest_row_at_pos(), but with structural checks"""
        # Some pre-flight checks
        src_path = self._drag_src_path
        if src_path is None:
            dest_treepath = None
            drop_pos = Gtk.TreeViewDropPosition.BEFORE
        root = self._docmodel.layer_stack
        assert len(root) > 0, "Unexpected row drag within an empty tree!"

        # Get GTK's purely position-based opinion, and decide what that
        # means within the real tree structure.
        dest_info = self.get_dest_row_at_pos(x, y)
        if dest_info is None:
            # GTK found no reference point. But it just hitboxes rows.
            # Therefore, for dropping, this indicates the big empty
            # space below all the layers.
            # Return the (nonexistent) path one below the end of the
            # root, and ask for an insert before that.
            dest_treepath = Gtk.TreePath([len(root)])
            drop_pos = Gtk.TreeViewDropPosition.BEFORE
        else:
            # GTK thinks it points at a reference point that actually
            # exists. Confirm that notion first...
            dest_treepath, drop_pos = dest_info
            dest_path = tuple(dest_treepath)
            dest_layer = root.deepget(dest_path)
            if dest_layer is None:
                dest_treepath = None
                drop_pos = Gtk.TreeViewDropPosition.BEFORE
            # Can't move a layer to its own position, or into itself,
            elif lib.layer.path_startswith(dest_path, src_path):
                dest_treepath = None
                drop_pos = Gtk.TreeViewDropPosition.BEFORE
            # or into any other layer that isn't a group.
            elif not isinstance(dest_layer, lib.layer.LayerStack):
                if drop_pos == Gtk.TreeViewDropPosition.INTO_OR_AFTER:
                    drop_pos = Gtk.TreeViewDropPosition.AFTER
                elif drop_pos == Gtk.TreeViewDropPosition.INTO_OR_BEFORE:
                    drop_pos = Gtk.TreeViewDropPosition.BEFORE

        if dest_treepath is not None:
            logger.debug(
                "Checked destination: %s %r",
                drop_pos.value_nick,
                tuple(dest_treepath),
            )
        return (dest_treepath, drop_pos)

    def _drag_motion_cb(self, view, context, x, y, t):
        dest_treepath, drop_pos = self._get_checked_dest_row_at_pos(x, y)
        self.set_drag_dest_row(dest_treepath, drop_pos)
        if dest_treepath is None:
            dest_path = None
            self._stop_hover_expand_timer()
        else:
            dest_path = tuple(dest_treepath)
        old_dest_path = self._drag_dest_path
        if old_dest_path != dest_path:
            self._drag_dest_path = dest_path
            if dest_path is not None:
                self._restart_hover_expand_timer(dest_path, x, y)
        return True

    def _restart_hover_expand_timer(self, path, x, y):
        self._stop_hover_expand_timer()
        root = self._docmodel.layer_stack
        layer = root.deepget(path)
        if not isinstance(layer, lib.layer.LayerStack):
            return
        if self.row_expanded(Gtk.TreePath(path)):
            return
        self._hover_expand_timer_id = GLib.timeout_add(
            int(self.DRAG_HOVER_EXPAND_TIME * 1000),
            self._hover_expand_timer_cb,
            path,
            x, y,
        )

    def _stop_hover_expand_timer(self):
        if self._hover_expand_timer_id is None:
            return
        GLib.source_remove(self._hover_expand_timer_id)
        self._hover_expand_timer_id = None

    def _hover_expand_timer_cb(self, path, x, y):
        self.expand_to_path(Gtk.TreePath(path))
        # The insertion marker may need updating after the expand
        dest_treepath, drop_pos = self._get_checked_dest_row_at_pos(x, y)
        self.set_drag_dest_row(dest_treepath, drop_pos)
        self._hover_expand_timer_id = None
        return False

    def _drag_leave_cb(self, view, context, t):
        """Reset the insertion point when the drag leaves"""
        logger.debug("drag-leave t=%d", t)
        self._stop_hover_expand_timer()
        self.set_drag_dest_row(None, Gtk.TreeViewDropPosition.BEFORE)

    def _get_insert_path_for_dest_row(self, dest_treepath, drop_pos):
        """Convert a GTK destination row to a tree insert point.

        This adjusts some path indices to be closer to what's intuitive
        at the end of the drag, based on what the user saw during it.
        The returned value must be checked before passing to the model
        to ensure it isn't the same as or within the dragged tree path.

        """
        root = self._docmodel.layer_stack
        if dest_treepath is None:
            n = len(root)
            return (n,)
        dest_path = tuple(dest_treepath)
        assert len(dest_path) > 0
        dest_layer = root.deepget(dest_path)
        gtvdp = Gtk.TreeViewDropPosition
        if isinstance(dest_layer, lib.layer.LayerStack):
            # Interpret Gtk's "into or before" as "into AND at the
            # start". Similar for "into or after".
            if drop_pos == gtvdp.INTO_OR_BEFORE:
                return tuple(list(dest_path) + [0])
            elif drop_pos == gtvdp.INTO_OR_AFTER:
                n = len(dest_layer)
                return tuple(list(dest_path) + [n])
        if drop_pos == gtvdp.BEFORE:
            return dest_path
        elif drop_pos == gtvdp.AFTER:
            is_expanded_group = (
                isinstance(dest_layer, lib.layer.LayerStack) and
                self.row_expanded(dest_treepath)
            )
            if is_expanded_group:
                # This highlights like an insert before its first item
                return tuple(list(dest_path) + [0])
            else:
                dest_path = list(dest_path)
                dest_path[-1] += 1
                return tuple(dest_path)
        else:
            raise NotImplemented("Unhandled position %r", drop_pos)

    def _drag_drop_cb(self, view, context, x, y, t):
        self._stop_hover_expand_timer()
        dest_treepath, drop_pos = self._get_checked_dest_row_at_pos(x, y)
        if dest_treepath is not None:
            src_path = self._drag_src_path
            dest_insert_path = self._get_insert_path_for_dest_row(
                dest_treepath,
                drop_pos,
            )
            if not lib.layer.path_startswith(dest_insert_path, src_path):
                logger.debug(
                    "drag-drop: move %r to insert at %r",
                    src_path,
                    dest_insert_path,
                )
                self._docmodel.restack_layer(src_path, dest_insert_path)
            Gtk.drag_finish(context, True, False, t)
            return True
        return False

    def _drag_end_cb(self, view, context):
        logger.debug("drag-end")
        self._stop_hover_expand_timer()
        self._drag_src_path = None
        self._drag_dest_path = None
        self.drag_ended()

    ## Model compat

    def do_drag_data_delete(self, context):
        """Suppress the default GtkWidgetClass.drag_data_delete handler.

        Suppress warning(s?) about missing default handlers, since our
        model no longer implements GtkTreeDragSource.

        """

    ## Model change tracking

    def _current_path_updated_cb(self, rootstack, layerpath):
        """Respond to the current layer changing in the doc-model"""
        self._update_selection()

    def _expand_layer_cb(self, rootstack, path):
        if not path:
            return
        treepath = Gtk.TreePath(path)
        self.expand_to_path(treepath)

    def _collapse_layer_cb(self, rootstack, path):
        if not path:
            return
        treepath = Gtk.TreePath(path)
        self.collapse_row(treepath)

    def _layer_content_changed_cb(self, rootstack, layer, *args):
        """Scroll to the current layer when it is modified."""
        if layer and layer is rootstack.current:
            self.scroll_to_current_layer()

    def _update_selection(self):
        sel = self.get_selection()
        root = self._docmodel.layer_stack
        layerpath = root.current_path
        if not layerpath:
            sel.unselect_all()
            return
        old_layerpath = None
        model, selected_paths = sel.get_selected_rows()
        if len(selected_paths) > 0:
            old_treepath = selected_paths[0]
            if old_treepath:
                old_layerpath = tuple(old_treepath.get_indices())
        if layerpath == old_layerpath:
            return
        sel.unselect_all()
        if len(layerpath) > 1:
            self.expand_to_path(Gtk.TreePath(layerpath[:-1]))
        if len(layerpath) > 0:
            sel.select_path(Gtk.TreePath(layerpath))
            self.scroll_to_current_layer()

    def scroll_to_current_layer(self, *_ignored):
        """Scroll to show the current layer"""
        sel = self.get_selection()
        tree_model, sel_row_paths = sel.get_selected_rows()
        if len(sel_row_paths) > 0:
            sel_row_path = sel_row_paths[0]
            if sel_row_path:
                self.scroll_to_cell(sel_row_path)

    ## Observable events (hook stuff here!)

    @event
    def current_layer_rename_requested(self):
        """Event: user double-clicked the name of the current layer"""

    @event
    def current_layer_changed(self):
        """Event: the current layer was just changed by clicking it"""

    @event
    def current_layer_menu_requested(self, gdkevent):
        """Event: user invoked the menu action over the current layer"""

    @event
    def drag_began(self):
        """Event: a drag has just started"""

    @event
    def drag_ended(self):
        """Event: a drag has just ended"""

    ## View datafuncs

    def _layer_visible_pixbuf_datafunc(self, column, cell, model, it, data):
        """Use an open/closed eye icon to show layer visibilities"""
        layer = model.get_layer(it=it)
        rootstack = model._root
        visible = True
        sensitive = not self._docmodel.layer_view_manager.current_view_locked
        if layer:
            # Layer visibility is based on the layer's natural hidden/
            # visible flag, but the layer stack can override that.
            if rootstack.current_layer_solo:
                visible = layer is rootstack.current
                sensitive = False
            else:
                visible = layer.visible
                sensitive = sensitive and layer.branch_visible

        icon_name = "mypaint-object-{}-symbolic".format(
            "visible" if visible else "hidden",
        )
        cell.set_property("icon-name", icon_name)
        cell.set_property("sensitive", sensitive)

    @staticmethod
    def _datafunc_get_pixbuf_height(initial, column, multiple=8, maximum=256):
        """Nearest multiple-of-n height for a pixbuf data cell."""
        ox, oy, w, h = column.cell_get_size(None)
        s = initial
        if h is not None:
            s = helpers.clamp((int(h // 8) * 8), s, maximum)
        return s

    def _layer_preview_pixbuf_datafunc(self, column, cell, model, it, data):
        """Render layer preview icons and type info."""

        # Get the layer's thumbnail
        layer = model.get_layer(it=it)
        thumb = layer.thumbnail

        # Scale it to a reasonable size for use as the preview.
        s = self._datafunc_get_pixbuf_height(32, column)
        preview = make_preview(thumb, s)
        cell.set_property("pixbuf", preview)

        # Add a watermark icon for non-painting layers.
        # Not completely sure this is a good idea...
        try:
            cache = self.__icon_cache
        except AttributeError:
            cache = {}
            self.__icon_cache = cache
        icon_name = layer.get_icon_name()
        icon_size = 16
        icon_size += 2   # allow fopr the outline
        icon = cache.get(icon_name, None)
        if not icon:
            icon = gui.drawutils.load_symbolic_icon(
                icon_name, icon_size,
                fg=(1, 1, 1, 1),
                outline=(0, 0, 0, 1),
            )
            cache[icon_name] = icon

        # Composite the watermark over the preview
        x = (preview.get_width() - icon_size) // 2
        y = (preview.get_height() - icon_size) // 2
        icon.composite(
            dest=preview,
            dest_x=x,
            dest_y=y,
            dest_width=icon_size,
            dest_height=icon_size,
            offset_x=x,
            offset_y=y,
            scale_x=1,
            scale_y=1,
            interp_type=GdkPixbuf.InterpType.NEAREST,
            overall_alpha=255/6,
        )

    @staticmethod
    def _layer_description_markup(layer):
        """GMarkup text description of a layer, used in the list."""
        name_markup = None
        description = None

        if layer is None:
            name_markup = escape(lib.layer.PlaceholderLayer.DEFAULT_NAME)
            description = C_(
                "Layers: description: no layer (\"never happens\" condition!)",
                u"?layer",
            )
        elif layer.name is None:
            name_markup = escape(layer.DEFAULT_NAME)
        else:
            name_markup = escape(layer.name)

        if layer is not None:
            desc_parts = []
            if isinstance(layer, lib.layer.LayerStack):
                name_markup = "<i>{}</i>".format(name_markup)

            # Mode (if it's interesting)
            if layer.mode in lib.modes.MODE_STRINGS:
                if layer.mode != lib.modes.default_mode():
                    s, d = lib.modes.MODE_STRINGS[layer.mode]
                    desc_parts.append(s)
            else:
                desc_parts.append(C_(
                    "Layers: description parts: unknown mode (fallback str!)",
                    u"?mode",
                ))

            # Visibility and opacity (if interesting)
            if not layer.visible:
                desc_parts.append(C_(
                    "Layers: description parts: layer hidden",
                    u"Hidden",
                ))
            elif layer.opacity < 1.0:
                desc_parts.append(C_(
                    "Layers: description parts: opacity percentage",
                    u"%d%% opaque" % (round(layer.opacity * 100),)
                ))

            # Locked flag (locked is interesting)
            if layer.locked:
                desc_parts.append(C_(
                    "Layers dockable: description parts: layer locked flag",
                    u"Locked",
                ))

            # Description of the layer's type.
            # Currently always used, for visual rhythm reasons, but it goes
            # on the end since it's perhaps the least interesting info.
            if layer.TYPE_DESCRIPTION is not None:
                desc_parts.append(layer.TYPE_DESCRIPTION)
            else:
                desc_parts.append(C_(
                    "Layers: description parts: unknown type (fallback str!)",
                    u"?type",
                ))

            # Stitch it all together
            if desc_parts:
                description = C_(
                    "Layers dockable: description parts joiner text",
                    u", ",
                ).join(desc_parts)
            else:
                description = None

        if description is None:
            markup_template = C_(
                "Layers dockable: markup for a layer with no description",
                u"{layer_name}",
            )
        else:
            markup_template = C_(
                "Layers dockable: markup for a layer with a description",
                '<span size="smaller">{layer_name}\n'
                '<span size="smaller">{layer_description}</span>'
                '</span>'
            )

        markup = markup_template.format(
            layer_name=name_markup,
            layer_description=escape(description),
        )
        return markup

    def _layer_name_text_datafunc(self, column, cell, model, it, data):
        """Show the layer name, with italics for layer groups"""
        layer = model.get_layer(it=it)
        markup = self._layer_description_markup(layer)

        attrs = Pango.AttrList()
        parse_result = Pango.parse_markup(markup, -1, '\000')
        parse_ok, attrs, text, accel_char = parse_result
        assert parse_ok
        cell.set_property("attributes", attrs)
        cell.set_property("text", text)

    @staticmethod
    def _get_layer_locked_icon_state(layer):
        icon_name = None
        sensitive = True
        if layer:
            locked = layer.locked
            sensitive = not layer.branch_locked
        if locked:
            icon_name = "mypaint-object-locked-symbolic"
        else:
            icon_name = "mypaint-object-unlocked-symbolic"
        return (icon_name, sensitive)

    def _layer_locked_pixbuf_datafunc(self, column, cell, model, it, data):
        """Use a padlock icon to show layer immutability statuses"""
        layer = model.get_layer(it=it)
        icon_name, sensitive = self._get_layer_locked_icon_state(layer)
        icon_visible = (icon_name is not None)
        cell.set_property("icon-name", icon_name)
        cell.set_visible(icon_visible)
        cell.set_property("sensitive", sensitive)

    ## Weird but necessary hacks

    def _post_show_cb(self, widget):
        # Ensure the tree selection matches the root stack's current layer.
        self._update_selection()

        # Match the flag column widths to the name column's height.
        # This only makes sense after the 1st text layout, sadly.
        GLib.idle_add(self._sizeify_flag_columns)

        return False

    def _sizeify_flag_columns(self):
        """Sneakily scale the fixed size of the flag icons to match texts.

        This can only be called after the list has rendered once, because
        GTK doesn't know how tall the treeview's rows will be till then.
        Therefore it's called in an idle callback after the first show.

        """
        # Get the maximum height for all columns.
        s = 0
        for col in self._columns:
            ox, oy, w, h = col.cell_get_size(None)
            if h > s:
                s = h
        if not s:
            return

        # Set that as the fixed size of the flag icon columns,
        # within reason, and force a re-layout.
        h = helpers.clamp(s, 24, 48)
        w = helpers.clamp(s, 24, 48)
        for col in [self._flags1_col, self._flags2_col]:
            for cell in col.get_cells():
                cell.set_fixed_size(w, h)
            col.set_min_width(w)
        for col in self._columns:
            col.queue_resize()


# Helper functions

def new_blend_mode_combo(modes, mode_strings):
    """Create and return a new blend mode combo box
    """
    store = Gtk.ListStore(int, str, bool, float)
    for mode in modes:
        label, desc = mode_strings.get(mode)
        sensitive = True
        scale = 1/1.2   # PANGO_SCALE_SMALL
        store.append([mode, label, sensitive, scale])
    combo = Gtk.ComboBox()
    combo.set_model(store)
    combo.set_hexpand(True)
    combo.set_vexpand(False)
    cell = Gtk.CellRendererText()
    combo.pack_start(cell, True)
    combo.add_attribute(cell, "text", 1)
    combo.add_attribute(cell, "sensitive", 2)
    combo.add_attribute(cell, "scale", 3)
    combo.set_wrap_width(2)
    combo.set_app_paintable(True)
    return combo

## Testing


def _test():
    """Test the custom model in an ad-hoc GUI window"""
    from lib.layer import PaintingLayer, LayerStack
    doc_model = Document()
    root = doc_model.layer_stack
    root.clear()
    layer_info = [
        ((0,), LayerStack(name="Layer 0")),
        ((0, 0), PaintingLayer(name="Layer 0:0")),
        ((0, 1), PaintingLayer(name="Layer 0:1")),
        ((0, 2), LayerStack(name="Layer 0:2")),
        ((0, 2, 0), PaintingLayer(name="Layer 0:2:0")),
        ((0, 2, 1), PaintingLayer(name="Layer 0:2:1")),
        ((0, 3), PaintingLayer(name="Layer 0:3")),
        ((1,), LayerStack(name="Layer 1")),
        ((1, 0), PaintingLayer(name="Layer 1:0")),
        ((1, 1), PaintingLayer(name="Layer 1:1")),
        ((1, 2), LayerStack(name="Layer 1:2")),
        ((1, 2, 0), PaintingLayer(name="Layer 1:2:0")),
        ((1, 2, 1), PaintingLayer(name="Layer 1:2:1")),
        ((1, 2, 2), PaintingLayer(name="Layer 1:2:2")),
        ((1, 2, 3), PaintingLayer(name="Layer 1:2:3")),
        ((1, 3), PaintingLayer(name="Layer 1:3")),
        ((1, 4), PaintingLayer(name="Layer 1:4")),
        ((1, 5), PaintingLayer(name="Layer 1:5")),
        ((1, 6), PaintingLayer(name="Layer 1:6")),
        ((2,), PaintingLayer(name="Layer 2")),
        ((3,), PaintingLayer(name="Layer 3")),
        ((4,), PaintingLayer(name="Layer 4")),
        ((5,), PaintingLayer(name="Layer 5")),
        ((6,), LayerStack(name="Layer 6")),
        ((6, 0), PaintingLayer(name="Layer 6:0")),
        ((6, 1), PaintingLayer(name="Layer 6:1")),
        ((6, 2), PaintingLayer(name="Layer 6:2")),
        ((6, 3), PaintingLayer(name="Layer 6:3")),
        ((6, 4), PaintingLayer(name="Layer 6:4")),
        ((6, 5), PaintingLayer(name="Layer 6:5")),
        ((7,), PaintingLayer(name="Layer 7")),
    ]
    for path, layer in layer_info:
        root.deepinsert(path, layer)
    root.set_current_path([4])

    icon_theme = Gtk.IconTheme.get_default()
    icon_theme.append_search_path("./desktop/icons")

    view = RootStackTreeView(doc_model)
    view_scroll = Gtk.ScrolledWindow()
    view_scroll.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
    scroll_pol = Gtk.PolicyType.AUTOMATIC
    view_scroll.set_policy(scroll_pol, scroll_pol)
    view_scroll.add(view)
    view_scroll.set_size_request(-1, 100)

    win = Gtk.Window()
    win.set_title(unicode(__package__))
    win.connect("destroy", Gtk.main_quit)
    win.add(view_scroll)
    win.set_default_size(300, 500)

    win.show_all()
    Gtk.main()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    _test()



# This file is part of MyPaint.
# Copyright (C) 2014 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Layer mode menu"""


## Imports

from __future__ import division, print_function

from lib.gibindings import Gtk

from lib.modes import STACK_MODES
from lib.modes import STANDARD_MODES
from lib.modes import MODE_STRINGS


## Class definitions

class LayerModeMenuItem (Gtk.ImageMenuItem):
    """Brush list menu item with a dynamic BrushGroupsMenu as its submenu

    This is instantiated by the app's UIManager using a FactoryAction which
    must be named "LayerMode" (see factoryaction.py).
    """

    __gtype_name__ = "MyPaintLayerModeMenuItem"

    def __init__(self):
        """Initialize, called by the LayerMode FactoryAction making a menu"""
        Gtk.ImageMenuItem.__init__(self)
        menu = Gtk.Menu()
        self._menu_items = []
        prev_item = None
        spec_separator = (None,)
        modes_menu_spec = (STACK_MODES + spec_separator + STANDARD_MODES)
        for mode in modes_menu_spec:
            if mode is None:
                menu.append(Gtk.SeparatorMenuItem())
                continue
            label, tooltip = MODE_STRINGS.get(mode)
            if prev_item is None:
                item = Gtk.RadioMenuItem()
            else:
                item = Gtk.RadioMenuItem(group=prev_item)
            item.set_label(label)
            item.set_tooltip_text(tooltip)
            item.connect("activate", self._item_activated_cb, mode)
            menu.append(item)
            self._menu_items.append((mode, item))
            prev_item = item
        self._submenu = menu
        self.set_submenu(self._submenu)
        self._submenu.show_all()
        from gui.application import get_app
        app = get_app()
        self._model = app.doc.model
        rootstack = self._model.layer_stack
        rootstack.layer_properties_changed += self._update_actions
        rootstack.current_path_updated += self._update_actions
        self._updating = False
        self._update_actions()

    def _item_activated_cb(self, item, mode):
        """Callback: Update the model when the user selects a menu item"""
        if self._updating:
            return
        self._model.set_current_layer_mode(mode)

    def _update_actions(self, *_ignored):
        """Updates menu actions to reflect the current layer's mode"""
        if self._updating:
            return
        self._updating = True
        rootstack = self._model.layer_stack
        current = rootstack.current
        for mode, item in self._menu_items:
            active = (mode == current.mode)
            if bool(item.get_active()) != active:
                item.set_active(active)
            item.set_sensitive(mode in current.PERMITTED_MODES)
        self._updating = False
